import uuid
from contextlib import contextmanager

from dagster_graphql.client.query import LAUNCH_PIPELINE_EXECUTION_MUTATION, SUBSCRIPTION_QUERY
from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector
from graphql import parse

from dagster import seven
from dagster.core.instance import DagsterInstance
from dagster.core.storage.pipeline_run import PipelineRunsFilter
from dagster.utils import file_relative_path, merge_dicts
from dagster.utils.test import get_temp_file_name

from .graphql_context_test_suite import ExecutingGraphQLContextTestMatrix
from .setup import csv_hello_world_solids_config
from .utils import sync_execute_get_run_log_data


class TestExecutePipeline(ExecutingGraphQLContextTestMatrix):
    def test_start_pipeline_execution(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config(),
                    'mode': 'default',
                }
            },
        )

        assert not result.errors
        assert result.data

        # just test existence
        assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        assert uuid.UUID(result.data['launchPipelineExecution']['run']['runId'])
        assert (
            result.data['launchPipelineExecution']['run']['pipeline']['name'] == 'csv_hello_world'
        )

    def test_basic_start_pipeline_execution_with_preset(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={'executionParams': {'selector': selector, 'preset': 'test_inline',}},
        )

        assert not result.errors
        assert result.data

        # just test existence
        assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        assert uuid.UUID(result.data['launchPipelineExecution']['run']['runId'])
        assert (
            result.data['launchPipelineExecution']['run']['pipeline']['name'] == 'csv_hello_world'
        )

    def test_basic_start_pipeline_execution_with_pipeline_def_tags(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'hello_world_with_tags')
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={'executionParams': {'selector': selector, 'mode': 'default',},},
        )

        assert not result.errors
        assert result.data['launchPipelineExecution']['run']['tags'] == [
            {'key': 'tag_key', 'value': 'tag_value'}
        ]

        # just test existence
        assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        assert uuid.UUID(result.data['launchPipelineExecution']['run']['runId'])
        assert (
            result.data['launchPipelineExecution']['run']['pipeline']['name']
            == 'hello_world_with_tags'
        )

    def test_basic_start_pipeline_execution_with_non_existent_preset(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')

        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={'executionParams': {'selector': selector, 'preset': 'undefined_preset',}},
        )

        assert not result.errors
        assert result.data
        assert result.data['launchPipelineExecution']['__typename'] == 'PresetNotFoundError'
        assert (
            result.data['launchPipelineExecution']['message']
            == 'Preset undefined_preset not found in pipeline csv_hello_world.'
        )

    def test_basic_start_pipeline_execution_with_preset_failure(self, graphql_context):
        subset_selector = infer_pipeline_selector(
            graphql_context, 'csv_hello_world', ['sum_sq_solid']
        )

        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={'executionParams': {'selector': subset_selector, 'preset': 'test_inline',}},
        )

        # while illegally defining selector.solid_selection
        assert not result.errors
        assert result.data
        assert (
            result.data['launchPipelineExecution']['__typename']
            == 'ConflictingExecutionParamsError'
        )
        assert (
            result.data['launchPipelineExecution']['message']
            == 'Invalid ExecutionParams. Cannot define selector.solid_selection when using a preset.'
        )

        # while illegally defining runConfigData
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'preset': 'test_inline',
                    'runConfigData': csv_hello_world_solids_config(),
                }
            },
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['launchPipelineExecution']['__typename']
            == 'ConflictingExecutionParamsError'
        )
        assert (
            result.data['launchPipelineExecution']['message']
            == 'Invalid ExecutionParams. Cannot define runConfigData when using a preset.'
        )

        # while illegally defining mode
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'preset': 'test_inline',
                    'mode': 'default',
                }
            },
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['launchPipelineExecution']['__typename']
            == 'ConflictingExecutionParamsError'
        )
        assert (
            result.data['launchPipelineExecution']['message']
            == 'Invalid ExecutionParams. Cannot define mode when using a preset.'
        )

    def test_basic_start_pipeline_execution_config_failure(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': {'solids': {'sum_solid': {'inputs': {'num': 384938439}}}},
                    'mode': 'default',
                }
            },
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['launchPipelineExecution']['__typename']
            == 'PipelineConfigValidationInvalid'
        )

    def test_basis_start_pipeline_not_found_error(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'sjkdfkdjkf')
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': {'solids': {'sum_solid': {'inputs': {'num': 'test.csv'}}}},
                    'mode': 'default',
                }
            },
        )

        assert not result.errors
        assert result.data

        # just test existence
        assert result.data['launchPipelineExecution']['__typename'] == 'PipelineNotFoundError'
        assert result.data['launchPipelineExecution']['pipelineName'] == 'sjkdfkdjkf'

    def test_basic_start_pipeline_execution_and_subscribe(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        run_logs = sync_execute_get_run_log_data(
            context=graphql_context,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': {
                        'solids': {
                            'sum_solid': {
                                'inputs': {'num': file_relative_path(__file__, '../data/num.csv')}
                            }
                        }
                    },
                    'mode': 'default',
                }
            },
        )

        assert run_logs['__typename'] == 'PipelineRunLogsSubscriptionSuccess'
        non_engine_event_types = [
            message['__typename']
            for message in run_logs['messages']
            if message['__typename'] != 'EngineEvent'
        ]
        expected_non_engine_event_types = [
            'PipelineStartEvent',
            'ExecutionStepStartEvent',
            'ExecutionStepInputEvent',
            'ExecutionStepOutputEvent',
            'ExecutionStepSuccessEvent',
            'ExecutionStepStartEvent',
            'ExecutionStepInputEvent',
            'ExecutionStepOutputEvent',
            'ExecutionStepSuccessEvent',
            'PipelineSuccessEvent',
        ]
        assert non_engine_event_types == expected_non_engine_event_types

    def test_subscription_query_error(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'naughty_programmer_pipeline')
        run_logs = sync_execute_get_run_log_data(
            context=graphql_context,
            variables={'executionParams': {'selector': selector, 'mode': 'default',}},
        )

        assert run_logs['__typename'] == 'PipelineRunLogsSubscriptionSuccess'

        step_run_log_entry = _get_step_run_log_entry(
            run_logs, 'throw_a_thing.compute', 'ExecutionStepFailureEvent'
        )

        assert step_run_log_entry
        # Confirm that it is the user stack

        assert step_run_log_entry['message'] == 'Execution of step "throw_a_thing.compute" failed.'
        assert step_run_log_entry['error']
        assert step_run_log_entry['level'] == 'ERROR'
        assert isinstance(step_run_log_entry['error']['stack'], list)

        assert 'bad programmer' in step_run_log_entry['error']['stack'][-1]

    def test_subscribe_bad_run_id(self, graphql_context):
        run_id = 'nope'
        subscription = execute_dagster_graphql(
            graphql_context, parse(SUBSCRIPTION_QUERY), variables={'runId': run_id}
        )

        subscribe_results = []
        subscription.subscribe(subscribe_results.append)

        assert len(subscribe_results) == 1
        subscribe_result = subscribe_results[0]

        assert (
            subscribe_result.data['pipelineRunLogs']['__typename']
            == 'PipelineRunLogsSubscriptionFailure'
        )
        assert subscribe_result.data['pipelineRunLogs']['missingRunId'] == 'nope'

    def test_basic_sync_execution_no_config(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'no_config_pipeline')
        result = sync_execute_get_run_log_data(
            context=graphql_context,
            variables={
                'executionParams': {'selector': selector, 'runConfigData': None, 'mode': 'default',}
            },
        )
        logs = result['messages']
        assert isinstance(logs, list)
        assert has_event_of_type(logs, 'PipelineStartEvent')
        assert has_event_of_type(logs, 'PipelineSuccessEvent')
        assert not has_event_of_type(logs, 'PipelineFailureEvent')

    def test_basic_inmemory_sync_execution(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        result = sync_execute_get_run_log_data(
            context=graphql_context,
            variables={
                'executionParams': {
                    'selector': selector,
                    'mode': 'default',
                    'runConfigData': csv_hello_world_solids_config(),
                }
            },
        )

        logs = result['messages']
        assert isinstance(logs, list)
        assert has_event_of_type(logs, 'PipelineStartEvent')
        assert has_event_of_type(logs, 'PipelineSuccessEvent')
        assert not has_event_of_type(logs, 'PipelineFailureEvent')

        assert first_event_of_type(logs, 'PipelineStartEvent')['level'] == 'DEBUG'

        sum_solid_output = get_step_output_event(logs, 'sum_solid.compute')
        assert sum_solid_output['stepKey'] == 'sum_solid.compute'

    def test_basic_filesystem_sync_execution(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        result = sync_execute_get_run_log_data(
            context=graphql_context,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': merge_dicts(
                        csv_hello_world_solids_config(), {'storage': {'filesystem': {}}}
                    ),
                    'mode': 'default',
                }
            },
        )

        logs = result['messages']
        assert isinstance(logs, list)
        assert has_event_of_type(logs, 'PipelineStartEvent')
        assert has_event_of_type(logs, 'PipelineSuccessEvent')
        assert not has_event_of_type(logs, 'PipelineFailureEvent')

        assert first_event_of_type(logs, 'PipelineStartEvent')['level'] == 'DEBUG'

        sum_solid_output = get_step_output_event(logs, 'sum_solid.compute')
        assert sum_solid_output['stepKey'] == 'sum_solid.compute'
        assert sum_solid_output['outputName'] == 'result'

    def test_basic_start_pipeline_execution_with_tags(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config(),
                    'executionMetadata': {
                        'tags': [{'key': 'dagster/test_key', 'value': 'test_value'}]
                    },
                    'mode': 'default',
                }
            },
        )

        assert not result.errors
        assert result.data
        assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'

        run = result.data['launchPipelineExecution']['run']
        run_id = run['runId']
        assert len(run['tags']) > 0
        assert any(
            [x['key'] == 'dagster/test_key' and x['value'] == 'test_value' for x in run['tags']]
        )

        # Check run storage
        runs_with_tag = graphql_context.instance.get_runs(
            filters=PipelineRunsFilter(tags={'dagster/test_key': 'test_value'})
        )
        assert len(runs_with_tag) == 1
        assert runs_with_tag[0].run_id == run_id

    def test_basic_start_pipeline_execution_with_materialization(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')

        with get_temp_file_name() as out_csv_path:

            run_config = {
                'solids': {
                    'sum_solid': {
                        'inputs': {'num': file_relative_path(__file__, '../data/num.csv')},
                        'outputs': [{'result': out_csv_path}],
                    }
                }
            }

            run_logs = sync_execute_get_run_log_data(
                context=graphql_context,
                variables={
                    'executionParams': {
                        'selector': selector,
                        'runConfigData': run_config,
                        'mode': 'default',
                    }
                },
            )

            step_mat_event = None

            for message in run_logs['messages']:
                if message['__typename'] == 'StepMaterializationEvent':
                    # ensure only one event
                    assert step_mat_event is None
                    step_mat_event = message

            # ensure only one event
            assert step_mat_event
            assert len(step_mat_event['materialization']['metadataEntries']) == 1
            assert step_mat_event['materialization']['metadataEntries'][0]['path'] == out_csv_path


@contextmanager
def sqlite_instance_with_manager_disabled():
    with seven.TemporaryDirectory() as temp_dir:
        yield DagsterInstance.local_temp(
            tempdir=temp_dir, overrides={'dagit': {'execution_manager': {'disabled': True}}}
        )


def _get_step_run_log_entry(pipeline_run_logs, step_key, typename):
    for message_data in pipeline_run_logs['messages']:
        if message_data['__typename'] == typename:
            if message_data['stepKey'] == step_key:
                return message_data


def first_event_of_type(logs, message_type):
    for log in logs:
        if log['__typename'] == message_type:
            return log
    return None


def has_event_of_type(logs, message_type):
    return first_event_of_type(logs, message_type) is not None


def get_step_output_event(logs, step_key, output_name='result'):
    for log in logs:
        if (
            log['__typename'] == 'ExecutionStepOutputEvent'
            and log['stepKey'] == step_key
            and log['outputName'] == output_name
        ):
            return log

    return None



ERROR_FRAGMENT = '''
fragment errorFragment on PythonError {
  message
  className
  stack
  cause {
    message
    className
    stack
    cause {
      message
      className
      stack
    }
  }
}
'''

STEP_EVENT_FRAGMENTS = (
    ERROR_FRAGMENT
    + '''
fragment metadataEntryFragment on EventMetadataEntry {
  __typename
  label
  description
  ... on EventFloatMetadataEntry {
    value
  }
  ... on EventJsonMetadataEntry {
    jsonString
  }
  ... on EventMarkdownMetadataEntry {
    mdStr
  }
  ... on EventPathMetadataEntry {
    path
  }
  ... on EventPythonArtifactMetadataEntry {
    module
    name
  }
  ... on EventTextMetadataEntry {
    text
  }
  ... on EventUrlMetadataEntry {
    url
  }
}

fragment stepEventFragment on StepEvent {
  stepKey
  solidHandleID
  ... on EngineEvent {
    metadataEntries {
      ...metadataEntryFragment
    }
    markerStart
    markerEnd
    engineError: error {
      ...errorFragment
    }
  }
  ... on ExecutionStepFailureEvent {
    error {
      ...errorFragment
    }
    failureMetadata {
      label
      description
      metadataEntries {
        ...metadataEntryFragment
      }
    }
  }
  ... on ExecutionStepInputEvent {
    inputName
    typeCheck {
      __typename
      success
      label
      description
      metadataEntries {
        ...metadataEntryFragment
      }
    }
  }
  ... on ExecutionStepOutputEvent {
    outputName
    typeCheck {
      __typename
      success
      label
      description
      metadataEntries {
        ...metadataEntryFragment
      }
    }
  }

  ... on ExecutionStepUpForRetryEvent {
    retryError: error {
      ...errorFragment
    }
    secondsToWait
  }

  ... on ObjectStoreOperationEvent {
        stepKey
        operationResult {
            op
            metadataEntries {
                ...metadataEntryFragment
            }
        }
    }

  ... on StepExpectationResultEvent {
    expectationResult {
      success
      label
      description
      metadataEntries {
        ...metadataEntryFragment
      }
    }
  }
  ... on StepMaterializationEvent {
    materialization {
      label
      description
      metadataEntries {
        ...metadataEntryFragment
      }
    }
  }

  ... on MessageEvent {
    runId
    message
    timestamp
    level
  }



}
'''
)

MESSAGE_EVENT_FRAGMENTS = (
    '''
fragment messageEventFragment on MessageEvent {
  runId
  message
  timestamp
  level
  ...stepEventFragment
  ... on PipelineInitFailureEvent {
    initError: error {
      ...errorFragment
    }
  }
}
'''
    + STEP_EVENT_FRAGMENTS
)

EXECUTE_RUN_IN_PROCESS_MUTATION = (
    ERROR_FRAGMENT
    + '''
mutation(
  $repositoryLocationName: String!
  $repositoryName: String!
  $runId: String!
) {
  executeRunInProcess(
    repositoryLocationName: $repositoryLocationName
    repositoryName: $repositoryName
    runId: $runId
  ) {
    __typename
    ... on ExecuteRunInProcessSuccess {
      run {
        runId
        status
        pipeline {
          name
        }

        tags {
          key
          value
        }
        runConfigYaml
        mode
      }
    }
    ... on InvalidStepError {
      invalidStepKey
    }
    ... on InvalidOutputError {
      stepKey
      invalidOutputName
    }
    ... on PipelineConfigValidationInvalid {
      pipelineName
      errors {
        __typename
        message
        path
        reason
      }
    }
    ... on PipelineNotFoundError {
      message
      pipelineName
    }
    ... on PythonError {
      ...errorFragment
    }
    ... on PipelineRunConflict {
      message
    }
    ... on PipelineRunNotFoundError {
      message
    }
    ... on ConflictingExecutionParamsError {
      message
    }
    ... on PresetNotFoundError {
      message
      preset
    }
  }
}
'''
)

EXECUTE_PLAN_MUTATION = (
    '''
mutation(
  $executionParams: ExecutionParams!
  $retries: Retries
) {
  executePlan(
    executionParams: $executionParams,
    retries: $retries,
  ) {
    __typename
    ... on InvalidStepError {
      invalidStepKey
    }
    ... on PipelineConfigValidationInvalid {
      pipelineName
      errors {
        __typename
        message
        path
        reason
      }
    }
    ... on PipelineNotFoundError {
      message
      pipelineName
    }
    ... on PythonError {
      message
      stack
    }
    ... on ExecutePlanSuccess {
      pipeline {
        name
      }
      hasFailures
      stepEvents {
        __typename
        ...stepEventFragment
      }
    }
  }
}
'''
    + STEP_EVENT_FRAGMENTS
)

RAW_EXECUTE_PLAN_MUTATION = '''
mutation(
  $executionParams: ExecutionParams!
  $retries: Retries
) {
  executePlan(
    executionParams: $executionParams,
    retries: $retries,
  ) {
    __typename
    ... on InvalidStepError {
      invalidStepKey
    }
    ... on PipelineConfigValidationInvalid {
      pipelineName
      errors {
        __typename
        message
        path
        reason
      }
    }
    ... on PipelineNotFoundError {
      message
      pipelineName
    }
    ... on PythonError {
      message
      stack
      cause {
          message
          stack
      }
    }
    ... on ExecutePlanSuccess {
      pipeline {
        name
      }
      hasFailures
      rawEventRecords
    }
  }
}
'''

SUBSCRIPTION_QUERY = (
    MESSAGE_EVENT_FRAGMENTS
    + '''
subscription subscribeTest($runId: ID!) {
  pipelineRunLogs(runId: $runId) {
    __typename
    ... on PipelineRunLogsSubscriptionSuccess {
      run {
        runId
      }
      messages {
        __typename
        ...messageEventFragment

        # only include here because unstable between runs
        ... on StepMaterializationEvent {
          materialization {
            label
            description
            metadataEntries {
              __typename
              ...metadataEntryFragment
            }
          }
        }

        ... on ExecutionStepFailureEvent {
          stepKey
          error {
            ...errorFragment
          }
        }
      }
    }

    ... on PipelineRunLogsSubscriptionFailure {
      missingRunId
      message
    }
  }
}

'''
)

LAUNCH_PIPELINE_EXECUTION_MUTATION = (
    ERROR_FRAGMENT
    + '''
mutation($executionParams: ExecutionParams!) {
  launchPipelineExecution(executionParams: $executionParams) {
    __typename

    ... on InvalidStepError {
      invalidStepKey
    }
    ... on InvalidOutputError {
      stepKey
      invalidOutputName
    }
    ... on LaunchPipelineRunSuccess {
      run {
        runId
        pipeline {
          name
        }
        tags {
          key
          value
        }
        status
        runConfigYaml
        mode
      }
    }
    ... on ConflictingExecutionParamsError {
      message
    }
    ... on PresetNotFoundError {
      preset
      message
    }
    ... on PipelineConfigValidationInvalid {
      pipelineName
      errors {
        __typename
        message
        path
        reason
      }
    }
    ... on PipelineNotFoundError {
      message
      pipelineName
    }
    ... on PythonError {
      ...errorFragment
    }
  }
}
'''
)


LAUNCH_PIPELINE_REEXECUTION_MUTATION = (
    ERROR_FRAGMENT
    + '''
mutation($executionParams: ExecutionParams!) {
  launchPipelineReexecution(executionParams: $executionParams) {
    __typename

    ... on PythonError {
      ...errorFragment
    }
    ... on LaunchPipelineRunSuccess {
      run {
        runId
        status
        pipeline {
          name
        }
        tags {
          key
          value
        }
        runConfigYaml
        mode
        rootRunId
        parentRunId
      }
    }
    ... on PipelineNotFoundError {
      message
      pipelineName
    }
    ... on PipelineConfigValidationInvalid {
      pipelineName
      errors {
        __typename
        message
        path
        reason
      }
    }
    ... on InvalidStepError {
      invalidStepKey
    }
    ... on InvalidOutputError {
      stepKey
      invalidOutputName
    }
    ... on ConflictingExecutionParamsError {
      message
    }
    ... on PresetNotFoundError {
      preset
      message
    }
  }
}
'''
)

PIPELINE_REEXECUTION_INFO_QUERY = '''
query ReexecutionInfoQuery($runId: ID!) {
  pipelineRunOrError(runId: $runId) {
    __typename
    ... on PipelineRun {
        stepKeysToExecute
      }
    }
  }
'''



from dagster import check
from dagster.core.definitions.events import AssetKey
from dagster.core.events import DagsterEventType


def get_assets(graphene_info):
    instance = graphene_info.context.instance
    if not instance.is_asset_aware:
        return graphene_info.schema.type_named('AssetsNotSupportedError')(
            message='The configured event log storage is not asset aware.'
        )

    asset_keys = instance.all_asset_keys()
    return graphene_info.schema.type_named('AssetConnection')(
        nodes=[graphene_info.schema.type_named('Asset')(key=asset_key) for asset_key in asset_keys]
    )


def get_asset(graphene_info, asset_key):
    check.inst_param(asset_key, 'asset_key', AssetKey)
    instance = graphene_info.context.instance
    if not instance.is_asset_aware:
        return graphene_info.schema.type_named('AssetsNotSupportedError')(
            message='The configured event log storage is not asset aware.'
        )

    return graphene_info.schema.type_named('Asset')(key=asset_key)


def get_asset_events(graphene_info, asset_key, cursor=None, limit=None):
    check.inst_param(asset_key, 'asset_key', AssetKey)
    check.opt_str_param(cursor, 'cursor')
    check.opt_int_param(limit, 'limit')
    instance = graphene_info.context.instance
    if not instance.is_asset_aware:
        return graphene_info.schema.type_named('AssetsNotSupportedError')(
            message='The configured event log storage is not asset aware.'
        )
    events = instance.events_for_asset_key(asset_key, cursor, limit)
    return [
        event
        for event in events
        if event.is_dagster_event
        and event.dagster_event.event_type_value == DagsterEventType.STEP_MATERIALIZATION.value
    ]


def get_asset_run_ids(graphene_info, asset_key):
    check.inst_param(asset_key, 'asset_key', AssetKey)
    instance = graphene_info.context.instance
    if not instance.is_asset_aware:
        return graphene_info.schema.type_named('AssetsNotSupportedError')(
            message='The configured event log storage is not asset aware.'
        )

    return instance.run_ids_for_asset_key(asset_key)


def get_assets_for_run_id(graphene_info, run_id):
    check.str_param(run_id, 'run_id')

    records = graphene_info.context.instance.all_logs(run_id)
    asset_keys = [
        record.dagster_event.asset_key
        for record in records
        if record.is_dagster_event and record.dagster_event.asset_key
    ]
    return [graphene_info.schema.type_named('Asset')(key=asset_key) for asset_key in asset_keys]



import graphql
import pytest
from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from .graphql_context_test_suite import ReadonlyGraphQLContextTestMatrix
from .utils import sync_execute_get_events

MODE_QUERY = '''
query ModesQuery($selector: PipelineSelector!, $mode: String!)
{
  runConfigSchemaOrError(selector: $selector, mode: $mode ) {
    __typename
    ... on RunConfigSchema {
      rootConfigType {
        key
        ... on CompositeConfigType {
          fields {
            configType {
              key
            }
          }
        }
      }
      allConfigTypes {
        key
      }
    }
  }
}
'''


def get_step_output(logs, step_key):
    for log in logs:
        if log['__typename'] == 'ExecutionStepOutputEvent' and log['stepKey'] == step_key:
            return log


def execute_modes_query(context, pipeline_name, mode):
    selector = infer_pipeline_selector(context, pipeline_name)
    return execute_dagster_graphql(
        context, MODE_QUERY, variables={'selector': selector, 'mode': mode,},
    )


def test_multi_mode_successful(graphql_context):
    selector = infer_pipeline_selector(graphql_context, 'multi_mode_with_resources')
    add_mode_logs = sync_execute_get_events(
        context=graphql_context,
        variables={
            'executionParams': {
                'selector': selector,
                'mode': 'add_mode',
                'runConfigData': {'resources': {'op': {'config': 2}}},
            }
        },
    )
    assert get_step_output(add_mode_logs, 'apply_to_three.compute')

    mult_mode_logs = sync_execute_get_events(
        context=graphql_context,
        variables={
            'executionParams': {
                'selector': selector,
                'mode': 'mult_mode',
                'runConfigData': {'resources': {'op': {'config': 2}}},
            }
        },
    )
    assert get_step_output(mult_mode_logs, 'apply_to_three.compute')

    double_adder_mode_logs = sync_execute_get_events(
        context=graphql_context,
        variables={
            'executionParams': {
                'selector': selector,
                'mode': 'double_adder',
                'runConfigData': {'resources': {'op': {'config': {'num_one': 2, 'num_two': 4}}}},
            }
        },
    )
    get_step_output(double_adder_mode_logs, 'apply_to_three.compute')


class TestModeDefinitions(ReadonlyGraphQLContextTestMatrix):
    def test_query_multi_mode(self, graphql_context):
        with pytest.raises(graphql.error.base.GraphQLError):
            execute_modes_query(graphql_context, 'multi_mode_with_resources', mode=None)

        modeful_result = execute_modes_query(
            graphql_context, 'multi_mode_with_resources', mode='add_mode'
        )
        assert modeful_result.data['runConfigSchemaOrError']['__typename'] == 'RunConfigSchema'



from dagster_graphql.client.query import LAUNCH_PIPELINE_EXECUTION_MUTATION, SUBSCRIPTION_QUERY
from dagster_graphql.implementation.context import DagsterGraphQLContext
from dagster_graphql.test.utils import execute_dagster_graphql

from dagster import check


def get_all_logs_for_finished_run_via_subscription(context, run_id):
    '''
    You should almost certainly ensure that this run has complete or terminated in order
    to get reliable results that you can test against.
    '''
    check.inst_param(context, 'context', DagsterGraphQLContext)

    run = context.instance.get_run_by_id(run_id)

    assert run.is_finished

    subscription = execute_dagster_graphql(context, SUBSCRIPTION_QUERY, variables={'runId': run_id})
    subscribe_results = []

    subscription.subscribe(subscribe_results.append)
    assert len(subscribe_results) == 1
    subscribe_result = subscribe_results[0]
    if subscribe_result.errors:
        raise Exception(subscribe_result.errors)
    assert not subscribe_result.errors
    assert subscribe_result.data

    # remove information that changes run-to-run
    assert 'pipelineRunLogs' in subscribe_result.data
    assert 'messages' in subscribe_result.data['pipelineRunLogs']
    for msg in subscribe_result.data['pipelineRunLogs']['messages']:
        msg['runId'] = '<runId dummy value>'
        msg['timestamp'] = '<timestamp dummy value>'

    return subscribe_result.data


def sync_execute_get_payload(variables, context):
    check.inst_param(context, 'context', DagsterGraphQLContext)

    result = execute_dagster_graphql(
        context, LAUNCH_PIPELINE_EXECUTION_MUTATION, variables=variables
    )

    assert result.data

    if result.data['launchPipelineExecution']['__typename'] != 'LaunchPipelineRunSuccess':
        raise Exception(result.data)

    context.drain_outstanding_executions()

    run_id = result.data['launchPipelineExecution']['run']['runId']
    return get_all_logs_for_finished_run_via_subscription(context, run_id)


def sync_execute_get_run_log_data(variables, context):
    check.inst_param(context, 'context', DagsterGraphQLContext)
    payload_data = sync_execute_get_payload(variables, context)
    return payload_data['pipelineRunLogs']


def sync_execute_get_events(variables, context):
    check.inst_param(context, 'context', DagsterGraphQLContext)
    return sync_execute_get_run_log_data(variables, context)['messages']



from dagster import EventMetadataEntry, check, seven
from dagster.core.definitions import (
    AssetMaterialization,
    ExpectationResult,
    Materialization,
    SolidHandle,
)
from dagster.core.definitions.events import PythonArtifactMetadataEntryData
from dagster.core.definitions.reconstructable import ReconstructableRepository
from dagster.core.events import (
    DagsterEvent,
    DagsterEventType,
    EngineEventData,
    StepExpectationResultData,
    StepMaterializationData,
)
from dagster.core.execution.plan.objects import (
    StepFailureData,
    StepInputData,
    StepOutputData,
    StepOutputHandle,
    StepRetryData,
    StepSuccessData,
    TypeCheckData,
    UserFailureData,
)
from dagster.utils.error import SerializableErrorInfo

HANDLED_EVENTS = {
    'ExecutionStepStartEvent': DagsterEventType.STEP_START,
    'ExecutionStepInputEvent': DagsterEventType.STEP_INPUT,
    'ExecutionStepOutputEvent': DagsterEventType.STEP_OUTPUT,
    'ExecutionStepFailureEvent': DagsterEventType.STEP_FAILURE,
    'ExecutionStepSkippedEvent': DagsterEventType.STEP_SKIPPED,
    'ExecutionStepSuccessEvent': DagsterEventType.STEP_SUCCESS,
    'StepMaterializationEvent': DagsterEventType.STEP_MATERIALIZATION,
    'StepExpectationResultEvent': DagsterEventType.STEP_EXPECTATION_RESULT,
    'ObjectStoreOperationEvent': DagsterEventType.OBJECT_STORE_OPERATION,
    'EngineEvent': DagsterEventType.ENGINE_EVENT,
    'ExecutionStepUpForRetryEvent': DagsterEventType.STEP_UP_FOR_RETRY,
    'ExecutionStepRestartEvent': DagsterEventType.STEP_RESTARTED,
}


def expectation_result_from_data(data):
    return ExpectationResult(
        success=data['success'],
        label=data['label'],
        description=data.get('description'),  # enforce?
        metadata_entries=list(event_metadata_entries(data.get('metadataEntries')) or []),
    )


def materialization_from_data(data):
    if data.get('asset_key'):
        return AssetMaterialization(
            asset_key=data['asset_key'],
            description=data.get('description'),
            metadata_entries=list(event_metadata_entries(data.get('metadataEntries')) or []),
        )
    return Materialization(
        label=data['label'],
        description=data.get('description'),  # enforce?
        metadata_entries=list(event_metadata_entries(data.get('metadataEntries')) or []),
    )


def error_from_data(data):
    return SerializableErrorInfo(
        message=data['message'],
        stack=data['stack'],
        cls_name=data['className'],
        cause=error_from_data(data['cause']) if data.get('cause') else None,
    )


def event_metadata_entries(metadata_entry_datas):
    if not metadata_entry_datas:
        return

    for metadata_entry_data in metadata_entry_datas:
        typename = metadata_entry_data['__typename']
        label = metadata_entry_data['label']
        description = metadata_entry_data.get('description')
        if typename == 'EventPathMetadataEntry':
            yield EventMetadataEntry.path(
                label=label, description=description, path=metadata_entry_data['path']
            )
        elif typename == 'EventJsonMetadataEntry':
            yield EventMetadataEntry.json(
                label=label,
                description=description,
                data=seven.json.loads(metadata_entry_data.get('jsonString', '')),
            )
        elif typename == 'EventMarkdownMetadataEntry':
            yield EventMetadataEntry.md(
                label=label, description=description, md_str=metadata_entry_data.get('md_str', '')
            )
        elif typename == 'EventTextMetadataEntry':
            yield EventMetadataEntry.text(
                label=label, description=description, text=metadata_entry_data['text']
            )
        elif typename == 'EventUrlMetadataEntry':
            yield EventMetadataEntry.url(
                label=label, description=description, url=metadata_entry_data['url']
            )
        elif typename == 'EventPythonArtifactMetadataEntry':
            yield EventMetadataEntry(
                label=label,
                description=description,
                entry_data=PythonArtifactMetadataEntryData(
                    metadata_entry_data['module'], metadata_entry_data['name']
                ),
            )
        else:
            check.not_implemented('TODO for type {}'.format(typename))


def dagster_event_from_dict(event_dict, pipeline_name):
    check.dict_param(event_dict, 'event_dict', key_type=str)
    check.str_param(pipeline_name, 'pipeline_name')

    # Get event_type
    event_type = HANDLED_EVENTS.get(event_dict['__typename'])
    if not event_type:
        raise Exception('unhandled event type %s' % event_dict['__typename'])

    # Get event_specific_data
    event_specific_data = None
    if event_type == DagsterEventType.STEP_OUTPUT:
        event_specific_data = StepOutputData(
            step_output_handle=StepOutputHandle(event_dict['stepKey'], event_dict['outputName']),
            type_check_data=TypeCheckData(
                success=event_dict['typeCheck']['success'],
                label=event_dict['typeCheck']['label'],
                description=event_dict.get('description'),
                metadata_entries=list(
                    event_metadata_entries(event_dict.get('metadataEntries')) or []
                ),
            ),
        )

    elif event_type == DagsterEventType.STEP_INPUT:
        event_specific_data = StepInputData(
            input_name=event_dict['inputName'],
            type_check_data=TypeCheckData(
                success=event_dict['typeCheck']['success'],
                label=event_dict['typeCheck']['label'],
                description=event_dict.get('description'),
                metadata_entries=list(
                    event_metadata_entries(event_dict.get('metadataEntries')) or []
                ),
            ),
        )
    elif event_type == DagsterEventType.STEP_SUCCESS:
        event_specific_data = StepSuccessData(0.0)

    elif event_type == DagsterEventType.STEP_UP_FOR_RETRY:
        event_specific_data = StepRetryData(
            error=error_from_data(event_dict['retryError']),
            seconds_to_wait=event_dict['secondsToWait'],
        )

    elif event_type == DagsterEventType.STEP_MATERIALIZATION:
        materialization = event_dict['materialization']
        event_specific_data = StepMaterializationData(
            materialization=materialization_from_data(materialization)
        )
    elif event_type == DagsterEventType.STEP_EXPECTATION_RESULT:
        expectation_result = expectation_result_from_data(event_dict['expectationResult'])
        event_specific_data = StepExpectationResultData(expectation_result)

    elif event_type == DagsterEventType.STEP_FAILURE:
        event_specific_data = StepFailureData(
            error_from_data(event_dict['error']),
            UserFailureData(
                label=event_dict['failureMetadata']['label'],
                description=event_dict['failureMetadata']['description'],
                metadata_entries=list(
                    event_metadata_entries(event_dict.get('metadataEntries')) or []
                ),
            )
            if event_dict.get('failureMetadata')
            else None,
        )

    elif event_type == DagsterEventType.ENGINE_EVENT:
        event_specific_data = EngineEventData(
            metadata_entries=list(event_metadata_entries(event_dict.get('metadataEntries'))),
            marker_start=event_dict.get('markerStart'),
            marker_end=event_dict.get('markerEnd'),
            error=error_from_data(event_dict['engineError'])
            if event_dict.get('engineError')
            else None,
        )

    return DagsterEvent(
        event_type_value=event_type.value,
        pipeline_name=pipeline_name,
        step_key=event_dict.get('stepKey'),
        solid_handle=SolidHandle.from_string(event_dict['solidHandleID'])
        if event_dict.get('solidHandleID')
        else None,
        # at the time of writing this:
        # * 'COMPUTE` is the only step kind
        # * this code should get deleted in the near future as we move away from
        #   dagster-graphql CLI as what we invoke in dask/k8s/etc.
        step_kind_value='COMPUTE' if event_dict.get('stepKey') else None,
        logging_tags=None,
        event_specific_data=event_specific_data,
    )


def construct_execute_plan_variables(
    recon_repo, mode, run_config, pipeline_name, run_id, step_keys
):
    check.inst_param(recon_repo, 'recon_repo', ReconstructableRepository)
    check.str_param(mode, 'mode')
    check.dict_param(run_config, 'run_config')
    check.str_param(pipeline_name, 'pipeline_name')
    check.str_param(run_id, 'run_id')
    check.list_param(step_keys, 'step_keys', of_type=str)

    variables = {
        'executionParams': {
            'runConfigData': run_config,
            'mode': mode,
            'selector': {
                'repositoryLocationName': '<<in_process>>',
                'repositoryName': recon_repo.get_definition().name,
                'pipelineName': pipeline_name,
            },
            'executionMetadata': {'runId': run_id},
            'stepKeys': step_keys,
        },
    }

    return variables


def parse_raw_log_lines(log_lines):
    '''Parses the raw log lines response from a dagster-graphql CLI invocation, typically from a
    Docker or Kubernetes container.

     - Log lines don't necessarily come back in order
     - Something else might log JSON
     - Docker appears to silently split very long log lines -- this is undocumented behavior

    Args:
        log_lines (List[str]): Log lines containing response JSON

    Returns:
        Dict: Parsed JSON response
    '''
    res = None
    lines = []
    coalesced = []
    in_split_line = False
    for line in log_lines:
        if not in_split_line and line.startswith('{'):
            if line.endswith('}'):
                lines.append(line)
                continue
            else:
                coalesced.append(line)
                in_split_line = True
                continue
        if in_split_line:
            coalesced.append(line)
            if line.endswith('}'):
                lines.append(''.join(coalesced))
                coalesced = []
                in_split_line = False

    for line in reversed(lines):
        try:
            res = seven.json.loads(line)
            break
        # If we don't get a GraphQL response, check the next line
        except seven.JSONDecodeError:
            continue

    return res



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots[
    'TestExecuteSchedule.test_tick_skip[sqlite_with_cli_api_run_launcher_in_process_env] 1'
] = {
    'stats': {'ticksFailed': 0, 'ticksSkipped': 1, 'ticksStarted': 0, 'ticksSucceeded': 0},
    'ticks': [{'status': 'SKIPPED', 'tickId': '1'}],
    'ticksCount': 1,
}

snapshots[
    'TestExecuteSchedule.test_tick_success[sqlite_with_cli_api_run_launcher_in_process_env] 1'
] = {
    'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 1},
    'ticks': [{'status': 'SUCCESS', 'tickId': '1'}],
    'ticksCount': 1,
}

snapshots[
    'TestExecuteSchedule.test_should_execute_scheduler_error[sqlite_with_cli_api_run_launcher_in_process_env] 1'
] = {
    'stats': {'ticksFailed': 1, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
    'ticks': [{'status': 'FAILURE', 'tickId': '1'}],
    'ticksCount': 1,
}

snapshots[
    'TestExecuteSchedule.test_tags_scheduler_error[sqlite_with_cli_api_run_launcher_in_process_env] 1'
] = {
    'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 1},
    'ticks': [{'status': 'SUCCESS', 'tickId': '1'}],
    'ticksCount': 1,
}

snapshots[
    'TestExecuteSchedule.test_run_config_scheduler_error[sqlite_with_cli_api_run_launcher_in_process_env] 1'
] = {
    'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 1},
    'ticks': [{'status': 'SUCCESS', 'tickId': '1'}],
    'ticksCount': 1,
}

snapshots[
    'TestExecuteSchedule.test_query_multiple_schedule_ticks[sqlite_with_cli_api_run_launcher_in_process_env] 1'
] = [
    {
        'name': 'dynamic_config',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'run_config_error_schedule',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 1},
            'ticks': [{'status': 'SUCCESS', 'tickId': '3'}],
            'ticksCount': 1,
        },
    },
    {
        'name': 'invalid_config_schedule',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'no_config_pipeline_hourly_schedule',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 1},
            'ticks': [{'status': 'SUCCESS', 'tickId': '1'}],
            'ticksCount': 1,
        },
    },
    {
        'name': 'no_config_pipeline_hourly_schedule_with_config_fn',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'no_config_should_execute',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 1, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [{'status': 'SKIPPED', 'tickId': '2'}],
            'ticksCount': 1,
        },
    },
    {
        'name': 'partition_based',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'partition_based_custom_selector',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'partition_based_decorator',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'partition_based_multi_mode_decorator',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'should_execute_error_schedule',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'solid_selection_daily_decorator',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'solid_selection_hourly_decorator',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'solid_selection_monthly_decorator',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'solid_selection_weekly_decorator',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'tagged_pipeline_override_schedule',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'tagged_pipeline_schedule',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
    {
        'name': 'tags_error_schedule',
        'scheduleState': {
            'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 0},
            'ticks': [],
            'ticksCount': 0,
        },
    },
]

snapshots[
    'TestExecuteSchedule.test_invalid_config_schedule_error[sqlite_with_cli_api_run_launcher_in_process_env] 1'
] = {
    'stats': {'ticksFailed': 0, 'ticksSkipped': 0, 'ticksStarted': 0, 'ticksSucceeded': 1},
    'ticks': [{'status': 'SUCCESS', 'tickId': '1'}],
    'ticksCount': 1,
}



from dagster_graphql.test.utils import (
    execute_dagster_graphql,
    infer_repository_selector,
    infer_schedule_selector,
)

from dagster.core.scheduler.scheduler import ScheduleStatus

from .setup import main_repo_location_name, main_repo_name

GET_SCHEDULE_STATES_QUERY = '''
query ScheduleStateQuery($repositorySelector: RepositorySelector!) {
  scheduleStatesOrError(repositorySelector: $repositorySelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on ScheduleStates {
      results {
        runs {
            runId
        }
        runsCount
        status
      }
    }
  }
}
'''

GET_SCHEDULE_STATES_WITHOUT_DEFINITIONS_QUERY = '''
query ScheduleStateQuery($repositorySelector: RepositorySelector!) {
  scheduleStatesOrError(repositorySelector: $repositorySelector, withNoScheduleDefinition: true) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on ScheduleStates {
      results {
        runs {
            runId
        }
        runsCount
        status
      }
    }
  }
}
'''

GET_SCHEDULE_DEFINITIONS_QUERY = '''
query ScheduleDefinitionsQuery($repositorySelector: RepositorySelector!) {
  scheduleDefinitionsOrError(repositorySelector: $repositorySelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on ScheduleDefinitions {
      results {
        name
        cronSchedule
        pipelineName
        solidSelection
        mode
        runConfigOrError {
          __typename
          ... on ScheduleRunConfig {
            yaml
          }
        }
      }
    }
  }
}
'''

GET_SCHEDULE_DEFINITION = '''
query getScheduleDefinition($scheduleSelector: ScheduleSelector!) {
  scheduleDefinitionOrError(scheduleSelector: $scheduleSelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on ScheduleDefinition {
      name
      partitionSet {
        name
      }
    }
  }
}
'''

RECONCILE_SCHEDULER_STATE_QUERY = '''
mutation(
  $repositorySelector: RepositorySelector!
) {
  reconcileSchedulerState(
    repositorySelector: $repositorySelector,
  ) {
      ... on PythonError {
        message
        stack
      }
      ... on ReconcileSchedulerStateSuccess {
        message
      }
    }
}
'''


START_SCHEDULES_QUERY = '''
mutation(
  $scheduleSelector: ScheduleSelector!
) {
  startSchedule(
    scheduleSelector: $scheduleSelector,
  ) {
    ... on PythonError {
      message
      className
      stack
    }
    ... on ScheduleStateResult {
      scheduleState {
        status
      }
    }
  }
}
'''


STOP_SCHEDULES_QUERY = '''
mutation(
  $scheduleSelector: ScheduleSelector!
) {
  stopRunningSchedule(
    scheduleSelector: $scheduleSelector,
  ) {
    ... on PythonError {
      message
      className
      stack
    }
    ... on ScheduleStateResult {
      scheduleState {
        status
      }
    }
  }
}
'''


def default_execution_params():
    return {
        "runConfigData": {"storage": {"filesystem": None}},
        "selector": {"name": "no_config_pipeline", "solidSelection": None},
        "mode": "default",
    }


def test_get_schedule_definitions_for_repository(graphql_context):
    selector = infer_repository_selector(graphql_context)
    result = execute_dagster_graphql(
        graphql_context, GET_SCHEDULE_DEFINITIONS_QUERY, variables={'repositorySelector': selector},
    )

    assert result.data
    assert result.data['scheduleDefinitionsOrError']
    assert result.data['scheduleDefinitionsOrError']['__typename'] == 'ScheduleDefinitions'

    external_repository = graphql_context.get_repository_location(
        main_repo_location_name()
    ).get_repository(main_repo_name())

    results = result.data['scheduleDefinitionsOrError']['results']
    assert len(results) == len(external_repository.get_external_schedules())

    for schedule in results:
        if schedule['name'] == 'run_config_error_schedule':
            assert schedule['runConfigOrError']['__typename'] == 'PythonError'
        elif schedule['name'] == 'invalid_config_schedule':
            assert (
                schedule['runConfigOrError']['yaml']
                == 'solids:\n  takes_an_enum:\n    config: invalid\n'
            )
        else:
            assert schedule['runConfigOrError']['yaml'] == 'storage:\n  filesystem: {}\n'


def test_get_schedule_states_for_repository(graphql_context):
    selector = infer_repository_selector(graphql_context)
    result = execute_dagster_graphql(
        graphql_context, GET_SCHEDULE_STATES_QUERY, variables={'repositorySelector': selector},
    )

    assert result.data
    assert result.data['scheduleStatesOrError']
    assert result.data['scheduleStatesOrError']['__typename'] == 'ScheduleStates'

    # Since we haven't run reconcile yet, there should be no states in storage
    results = result.data['scheduleStatesOrError']['results']
    assert len(results) == 0


def test_get_schedule_state_with_for_repository_not_reconciled(graphql_context):
    selector = infer_repository_selector(graphql_context)
    result = execute_dagster_graphql(
        graphql_context, GET_SCHEDULE_STATES_QUERY, variables={'repositorySelector': selector},
    )

    assert result.data
    assert result.data['scheduleStatesOrError']
    assert result.data['scheduleStatesOrError']['__typename'] == 'ScheduleStates'

    # Since we haven't run reconcile yet, there should be no states in storage
    results = result.data['scheduleStatesOrError']['results']
    assert len(results) == 0


def test_get_schedule_states_for_repository_after_reconcile(graphql_context):
    selector = infer_repository_selector(graphql_context)

    external_repository = graphql_context.get_repository_location(
        main_repo_location_name()
    ).get_repository(main_repo_name())
    graphql_context.instance.reconcile_scheduler_state(external_repository)

    result = execute_dagster_graphql(
        graphql_context, GET_SCHEDULE_STATES_QUERY, variables={'repositorySelector': selector},
    )

    assert result.data
    assert result.data['scheduleStatesOrError']
    assert result.data['scheduleStatesOrError']['__typename'] == 'ScheduleStates'

    results = result.data['scheduleStatesOrError']['results']
    assert len(results) == len(external_repository.get_external_schedules())

    for schedule_state in results:
        assert schedule_state['status'] == ScheduleStatus.STOPPED.value


def test_get_schedule_states_for_repository_after_reconcile_using_mutation(graphql_context):
    selector = infer_repository_selector(graphql_context)

    external_repository = graphql_context.get_repository_location(
        main_repo_location_name()
    ).get_repository(main_repo_name())

    result = execute_dagster_graphql(
        graphql_context,
        RECONCILE_SCHEDULER_STATE_QUERY,
        variables={'repositorySelector': selector},
    )

    assert result.data
    assert result.data['reconcileSchedulerState']
    assert result.data['reconcileSchedulerState']['message'] == "Success"

    result = execute_dagster_graphql(
        graphql_context, GET_SCHEDULE_STATES_QUERY, variables={'repositorySelector': selector},
    )

    assert result.data
    assert result.data['scheduleStatesOrError']
    assert result.data['scheduleStatesOrError']['__typename'] == 'ScheduleStates'

    results = result.data['scheduleStatesOrError']['results']
    assert len(results) == len(external_repository.get_external_schedules())

    for schedule_state in results:
        assert schedule_state['status'] == ScheduleStatus.STOPPED.value


def test_get_schedule_states_for_repository_with_removed_schedule_definitions(graphql_context):
    selector = infer_repository_selector(graphql_context)

    external_repository = graphql_context.get_repository_location(
        main_repo_location_name()
    ).get_repository(main_repo_name())
    graphql_context.instance.reconcile_scheduler_state(external_repository)

    result = execute_dagster_graphql(
        graphql_context,
        GET_SCHEDULE_STATES_WITHOUT_DEFINITIONS_QUERY,
        variables={'repositorySelector': selector},
    )

    assert result.data['scheduleStatesOrError']
    assert result.data['scheduleStatesOrError']['__typename'] == 'ScheduleStates'
    results = result.data['scheduleStatesOrError']['results']
    assert len(results) == 0


def test_start_and_stop_schedule(graphql_context):
    # selector = infer_repository_selector(graphql_context)

    external_repository = graphql_context.get_repository_location(
        main_repo_location_name()
    ).get_repository(main_repo_name())
    graphql_context.instance.reconcile_scheduler_state(external_repository)

    schedule_selector = infer_schedule_selector(
        graphql_context, 'no_config_pipeline_hourly_schedule'
    )

    # Start a single schedule
    start_result = execute_dagster_graphql(
        graphql_context, START_SCHEDULES_QUERY, variables={'scheduleSelector': schedule_selector},
    )
    assert (
        start_result.data['startSchedule']['scheduleState']['status']
        == ScheduleStatus.RUNNING.value
    )

    # Stop a single schedule
    stop_result = execute_dagster_graphql(
        graphql_context, STOP_SCHEDULES_QUERY, variables={'scheduleSelector': schedule_selector},
    )
    assert (
        stop_result.data['stopRunningSchedule']['scheduleState']['status']
        == ScheduleStatus.STOPPED.value
    )


def test_get_single_schedule_definition(graphql_context):
    context = graphql_context
    instance = context.instance

    instance.reconcile_scheduler_state(
        external_repository=context.get_repository_location(
            main_repo_location_name()
        ).get_repository(main_repo_name()),
    )

    schedule_selector = infer_schedule_selector(context, 'partition_based_multi_mode_decorator')
    result = execute_dagster_graphql(
        context, GET_SCHEDULE_DEFINITION, variables={'scheduleSelector': schedule_selector}
    )

    assert result.data
    assert result.data['scheduleDefinitionOrError']['__typename'] == 'ScheduleDefinition'
    assert result.data['scheduleDefinitionOrError']['partitionSet']



COMPOSITES_QUERY = '''
query CompositesQuery($selector: PipelineSelector!) {
  pipelineOrError(params: $selector) {
    __typename
    ... on Pipeline {
      name
      solidHandles {
        handleID
        solid {
          ...SolidInfo
        }
      }
    }
    ... on PythonError {
      message
      stack
    }
  }
}

fragment SolidInfo on Solid {
  name
  inputs {
    definition {
      name
    }
    dependsOn {
      solid {
        name
      }
    }
  }
  outputs {
    definition {
      name
    }
    dependedBy {
      solid {
        name
      }
    }
  }
  definition {
    ... on CompositeSolidDefinition {
      solids { name }
      inputMappings {
        definition { name }
        mappedInput {
          definition { name }
          solid { name }
        }
      }
      outputMappings {
        definition {
          name
        }
        mappedOutput {
          definition { name }
          solid { name }
        }
      }
    }
  }
}
'''

PARENT_ID_QUERY = '''
query withParent($selector: PipelineSelector!, $parentHandleID: String) {
  pipelineOrError(params: $selector) {
    __typename
    ... on Pipeline {
      name
      solidHandles(parentHandleID: $parentHandleID) {
        handleID
      }
    }
  }
}
'''

SOLID_ID_QUERY = '''
query solidFetch($selector: PipelineSelector!, $id: String!) {
  pipelineOrError(params: $selector) {
    __typename
    ... on Pipeline {
      name
      solidHandle(handleID: $id) {
        handleID
      }
    }
  }
}
'''

COMPOSITES_QUERY_NESTED_DEPENDS_ON_DEPENDS_BY_CORE = '''
query CompositesQuery($selector: PipelineSelector!) {
  pipelineOrError(params: $selector) {
    __typename
    ... on Pipeline {
      name
      solidHandles {
        handleID
        solid {
          ...SolidInfo
        }
      }
    }
  }
}
'''


NESTED_INPUT_DEPENDS_ON = '''
fragment SolidInfo on Solid {
  outputs {
    dependedBy {
      solid {
        name
        inputs {
          definition { name }
          dependsOn {
            definition {
              name
            }
          }
        }
      }
    }
  }
}
'''

NESTED_OUTPUT_DEPENDED_BY = '''
fragment SolidInfo on Solid {
  name
  inputs {
    definition {
      name
    }
    dependsOn {
      solid {
        name
        outputs {
          dependedBy {
            definition { name }
          }
        }
      }
    }
  }
}
'''



from dagster_graphql.test.utils import infer_pipeline_selector

from .graphql_context_test_suite import ExecutingGraphQLContextTestMatrix
from .utils import sync_execute_get_events


class TestMaterializations(ExecutingGraphQLContextTestMatrix):
    def test_materializations(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, 'materialization_pipeline')
        logs = sync_execute_get_events(
            context=graphql_context,
            variables={'executionParams': {'selector': selector, 'mode': 'default',}},
        )

        materializations = [log for log in logs if log['__typename'] == 'StepMaterializationEvent']
        assert len(materializations) == 1
        mat = materializations[0]['materialization']
        assert mat['label'] == 'all_types'

        text_entry = mat['metadataEntries'][0]
        assert text_entry['__typename'] == 'EventTextMetadataEntry'
        assert text_entry['text']

        text_entry = mat['metadataEntries'][1]
        assert text_entry['__typename'] == 'EventUrlMetadataEntry'
        assert text_entry['url']

        text_entry = mat['metadataEntries'][2]
        assert text_entry['__typename'] == 'EventPathMetadataEntry'
        assert text_entry['path']

        text_entry = mat['metadataEntries'][3]
        assert text_entry['__typename'] == 'EventJsonMetadataEntry'
        assert text_entry['jsonString']

        text_entry = mat['metadataEntries'][4]
        assert text_entry['__typename'] == 'EventPythonArtifactMetadataEntry'
        assert text_entry['module']
        assert text_entry['name']

        text_entry = mat['metadataEntries'][5]
        assert text_entry['__typename'] == 'EventPythonArtifactMetadataEntry'
        assert text_entry['module']
        assert text_entry['name']

        text_entry = mat['metadataEntries'][6]
        assert text_entry['__typename'] == 'EventFloatMetadataEntry'
        assert text_entry['value']

        non_engine_event_logs = [
            message for message in logs if message['__typename'] != 'EngineEvent'
        ]

        snapshot.assert_match([message['__typename'] for message in non_engine_event_logs])



from dagster_graphql.client.query import EXECUTE_RUN_IN_PROCESS_MUTATION
from dagster_graphql.test.utils import execute_dagster_graphql

from dagster import DagsterInstance
from dagster.core.host_representation.handle import IN_PROCESS_NAME
from dagster.core.utils import make_new_run_id

from .graphql_context_test_suite import ExecutingGraphQLContextTestMatrix
from .setup import (
    csv_hello_world,
    csv_hello_world_solids_config,
    define_test_out_of_process_context,
    main_repo_location_name,
    main_repo_name,
)


class TestStartPipelineForCreatedRunInHostedUserProcess(ExecutingGraphQLContextTestMatrix):
    def test_synchronously_execute_run_within_hosted_user_process(self, graphql_context):
        pipeline_run = graphql_context.instance.create_run_for_pipeline(
            pipeline_def=csv_hello_world, run_config=csv_hello_world_solids_config()
        )

        result = execute_dagster_graphql(
            graphql_context,
            EXECUTE_RUN_IN_PROCESS_MUTATION,
            variables={
                'runId': pipeline_run.run_id,
                'repositoryLocationName': main_repo_location_name(),
                'repositoryName': main_repo_name(),
            },
        )
        assert result.data
        assert result.data['executeRunInProcess']['__typename'] == 'ExecuteRunInProcessSuccess'

    def test_synchronously_execute_run_within_hosted_user_process_not_found(self, graphql_context):
        run_id = make_new_run_id()
        result = execute_dagster_graphql(
            graphql_context,
            EXECUTE_RUN_IN_PROCESS_MUTATION,
            variables={
                'runId': run_id,
                'repositoryLocationName': main_repo_location_name(),
                'repositoryName': main_repo_name(),
            },
        )

        assert result.data
        assert result.data['executeRunInProcess']['__typename'] == 'PipelineRunNotFoundError'


# should be removed as https://github.com/dagster-io/dagster/issues/2608 is completed
def test_shameful_workaround():
    graphql_context = define_test_out_of_process_context(DagsterInstance.ephemeral())

    pipeline_run = graphql_context.instance.create_run_for_pipeline(
        pipeline_def=csv_hello_world, run_config=csv_hello_world_solids_config()
    )

    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_RUN_IN_PROCESS_MUTATION,
        variables={
            'runId': pipeline_run.run_id,
            # in corect in process name represents launching from user process
            'repositoryLocationName': IN_PROCESS_NAME,
            'repositoryName': main_repo_name(),
        },
    )
    assert result.data
    assert result.data['executeRunInProcess']['__typename'] == 'ExecuteRunInProcessSuccess'

    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_RUN_IN_PROCESS_MUTATION,
        variables={
            'runId': pipeline_run.run_id,
            # but we don't apply workaround to other names
            'repositoryLocationName': 'some_other_name',
            'repositoryName': main_repo_name(),
        },
    )
    assert result.data
    assert result.data['executeRunInProcess']['__typename'] == 'PipelineNotFoundError'



from __future__ import absolute_import

from dagster_graphql import dauphin
from dagster_graphql.implementation.fetch_solids import get_solid, get_solids

from dagster import check
from dagster.core.host_representation import (
    ExternalRepository,
    PythonEnvRepositoryLocationHandle,
    RepositoryLocation,
)
from dagster.core.origin import RepositoryOrigin


class DauphinRepository(dauphin.ObjectType):
    class Meta(object):
        name = 'Repository'

    def __init__(self, repository, repository_location):
        self._repository = check.inst_param(repository, 'repository', ExternalRepository)
        self._repository_location = check.inst_param(
            repository_location, 'repository_location', RepositoryLocation
        )
        super(DauphinRepository, self).__init__(name=repository.name)

    id = dauphin.NonNull(dauphin.ID)
    name = dauphin.NonNull(dauphin.String)
    location = dauphin.NonNull('RepositoryLocation')
    pipelines = dauphin.non_null_list('Pipeline')
    usedSolids = dauphin.Field(dauphin.non_null_list('UsedSolid'))
    usedSolid = dauphin.Field('UsedSolid', name=dauphin.NonNull(dauphin.String))
    origin = dauphin.NonNull('RepositoryOrigin')

    def resolve_id(self, _graphene_info):
        return self._repository.get_origin_id()

    def resolve_origin(self, graphene_info):
        return graphene_info.schema.type_named('RepositoryOrigin')(self._repository.get_origin())

    def resolve_location(self, graphene_info):
        return graphene_info.schema.type_named('RepositoryLocation')(self._repository_location)

    def resolve_pipelines(self, graphene_info):
        return sorted(
            [
                graphene_info.schema.type_named('Pipeline')(pipeline)
                for pipeline in self._repository.get_all_external_pipelines()
            ],
            key=lambda pipeline: pipeline.name,
        )

    def resolve_usedSolid(self, _graphene_info, name):
        return get_solid(self._repository, name)

    def resolve_usedSolids(self, _graphene_info):
        return get_solids(self._repository)


class DauphinRepositoryOrigin(dauphin.ObjectType):
    class Meta(object):
        name = 'RepositoryOrigin'

    executable_path = dauphin.NonNull(dauphin.String)
    code_pointer_description = dauphin.NonNull(dauphin.String)

    def __init__(self, origin):
        self._origin = check.inst_param(origin, 'origin', RepositoryOrigin)

    def resolve_executable_path(self, _graphene_info):
        return self._origin.executable_path

    def resolve_code_pointer_description(self, _graphene_info):
        return self._origin.code_pointer.describe()


class DauphinRepositoryLocation(dauphin.ObjectType):
    class Meta(object):
        name = 'RepositoryLocation'

    name = dauphin.NonNull(dauphin.String)
    is_reload_supported = dauphin.NonNull(dauphin.Boolean)
    environment_path = dauphin.String()
    repositories = dauphin.non_null_list('Repository')

    def __init__(self, location):
        self._location = check.inst_param(location, 'location', RepositoryLocation)
        environment_path = (
            location.location_handle.executable_path
            if isinstance(location.location_handle, PythonEnvRepositoryLocationHandle)
            else None
        )

        check.invariant(location.name is not None)

        super(DauphinRepositoryLocation, self).__init__(
            name=location.name,
            environment_path=environment_path,
            is_reload_supported=location.is_reload_supported,
        )

    def resolve_repositories(self, graphene_info):
        return [
            graphene_info.schema.type_named('Repository')(repository, self._location)
            for repository in self._location.get_repositories().values()
        ]


class DauphinRepositoryConnection(dauphin.ObjectType):
    class Meta(object):
        name = 'RepositoryConnection'

    nodes = dauphin.non_null_list('Repository')



from dagster_graphql import dauphin


class DauphinCursor(dauphin.Int, dauphin.Scalar):
    class Meta(object):
        name = 'Cursor'


class DauphinPageInfo(dauphin.ObjectType):
    class Meta(object):
        name = 'PageInfo'

    lastCursor = dauphin.Field('Cursor')
    hasNextPage = dauphin.Field(dauphin.Boolean)
    hasPreviousPage = dauphin.Field(dauphin.Boolean)
    count = dauphin.NonNull(dauphin.Int)
    totalCount = dauphin.NonNull(dauphin.Int)



import re

from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from dagster import check
from dagster.core.execution.plan.objects import StepOutputHandle
from dagster.core.storage.intermediate_store import build_fs_intermediate_store
from dagster.core.storage.intermediates_manager import IntermediateStoreIntermediatesManager
from dagster.utils import file_relative_path, merge_dicts
from dagster.utils.test import get_temp_file_name

from .graphql_context_test_suite import ReadonlyGraphQLContextTestMatrix
from .setup import (
    PoorMansDataFrame,
    csv_hello_world,
    csv_hello_world_solids_config,
    csv_hello_world_solids_config_fs_storage,
)

EXECUTION_PLAN_QUERY = '''
query PipelineQuery($runConfigData: RunConfigData, $pipeline: PipelineSelector!, $mode: String!) {
  executionPlanOrError(runConfigData: $runConfigData, pipeline: $pipeline, mode: $mode) {
    __typename
    ... on ExecutionPlan {
      steps {
        key
        solidHandleID
        kind
        inputs {
          name
          type {
            name
          }
          dependsOn {
            key
          }
        }
        outputs {
          name
          type {
            name
          }
        }
        metadata {
          key
          value
        }
      }
    }
    ... on PipelineNotFoundError {
        pipelineName
    }
    ... on PipelineConfigValidationInvalid {
        pipelineName
        errors { message }
    }
    ... on PythonError {
        message
        stack
        cause {
            message
            stack
        }
    }
  }
}
'''

EXECUTE_PLAN_QUERY = '''
mutation ($executionParams: ExecutionParams!, $retries: Retries) {
    executePlan(executionParams: $executionParams, retries: $retries) {
        __typename
        ... on ExecutePlanSuccess {
            pipeline { name }
            hasFailures
            stepEvents {
                __typename
                ... on MessageEvent {
                    message
                }
                stepKey
                ... on ExecutionStepOutputEvent {
                    outputName
                }
                ... on StepMaterializationEvent {
                    materialization {
                        label
                        description
                        metadataEntries {
                            ... on EventPathMetadataEntry {
                                label
                                description
                                path
                            }
                        }
                    }
                }
                ... on ExecutionStepFailureEvent {
                    error {
                        message
                    }
                }
            }
        }
        ... on PipelineConfigValidationInvalid {
            pipelineName
            errors { message }
        }
        ... on PipelineNotFoundError {
            pipelineName
        }
        ... on PythonError {
            message
            stack
            cause {
                message
                stack
            }
        }
    }
}
'''


def get_nameset(llist):
    return {item['name'] for item in llist}


def get_named_thing(llist, name):
    for cn in llist:
        if cn['name'] == name:
            return cn

    check.failed('not found')


# for snapshot testing remove any varying values
def clean_log_messages(result_data):
    for idx in range(len(result_data['executePlan']['stepEvents'])):
        message = result_data['executePlan']['stepEvents'][idx].get('message')
        if message is not None:
            result_data['executePlan']['stepEvents'][idx]['message'] = re.sub(
                r'(\d+(\.\d+)?)', '{N}', message
            )
    return result_data


def test_success_whole_execution_plan(graphql_context, snapshot):
    run_config = csv_hello_world_solids_config_fs_storage()
    pipeline_run = graphql_context.instance.create_run_for_pipeline(
        pipeline_def=csv_hello_world, run_config=run_config
    )
    selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            'executionParams': {
                'selector': selector,
                'runConfigData': run_config,
                'stepKeys': None,
                'executionMetadata': {'runId': pipeline_run.run_id},
                'mode': 'default',
            },
        },
    )

    query_result = result.data['executePlan']
    assert query_result['__typename'] == 'ExecutePlanSuccess'
    assert query_result['pipeline']['name'] == 'csv_hello_world'
    assert query_result['hasFailures'] is False
    step_events = {
        step_event['stepKey']: step_event
        for step_event in query_result['stepEvents']
        if step_event['stepKey']
    }
    assert 'sum_solid.compute' in step_events
    assert 'sum_sq_solid.compute' in step_events

    snapshot.assert_match(clean_log_messages(result.data))
    store = build_fs_intermediate_store(
        graphql_context.instance.intermediates_directory, pipeline_run.run_id
    )
    intermediates_manager = IntermediateStoreIntermediatesManager(store)
    assert intermediates_manager.has_intermediate(None, StepOutputHandle('sum_solid.compute'))
    assert intermediates_manager.has_intermediate(None, StepOutputHandle('sum_sq_solid.compute'))


def test_success_whole_execution_plan_with_filesystem_config(graphql_context, snapshot):
    instance = graphql_context.instance
    selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
    run_config = merge_dicts(csv_hello_world_solids_config(), {'storage': {'filesystem': {}}})
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=csv_hello_world, run_config=run_config
    )
    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            'executionParams': {
                'selector': selector,
                'runConfigData': run_config,
                'stepKeys': None,
                'executionMetadata': {'runId': pipeline_run.run_id},
                'mode': 'default',
            },
        },
    )

    query_result = result.data['executePlan']

    assert query_result['__typename'] == 'ExecutePlanSuccess'
    assert query_result['pipeline']['name'] == 'csv_hello_world'
    assert query_result['hasFailures'] is False
    step_events = {
        step_event['stepKey']: step_event
        for step_event in query_result['stepEvents']
        if step_event['stepKey']
    }
    assert 'sum_solid.compute' in step_events
    assert 'sum_sq_solid.compute' in step_events

    snapshot.assert_match(clean_log_messages(result.data))
    store = build_fs_intermediate_store(instance.intermediates_directory, pipeline_run.run_id)
    intermediates_manager = IntermediateStoreIntermediatesManager(store)
    assert intermediates_manager.has_intermediate(None, StepOutputHandle('sum_solid.compute'))
    assert intermediates_manager.has_intermediate(None, StepOutputHandle('sum_sq_solid.compute'))


def test_success_whole_execution_plan_with_in_memory_config(graphql_context, snapshot):
    instance = graphql_context.instance
    selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
    run_config = merge_dicts(csv_hello_world_solids_config(), {'storage': {'in_memory': {}}})
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=csv_hello_world, run_config=run_config
    )
    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            'executionParams': {
                'selector': selector,
                'runConfigData': run_config,
                'stepKeys': None,
                'executionMetadata': {'runId': pipeline_run.run_id},
                'mode': 'default',
            },
        },
    )

    query_result = result.data['executePlan']

    assert query_result['__typename'] == 'ExecutePlanSuccess'
    assert query_result['pipeline']['name'] == 'csv_hello_world'
    assert query_result['hasFailures'] is False
    step_events = {
        step_event['stepKey']: step_event
        for step_event in query_result['stepEvents']
        if step_event['stepKey']
    }
    assert 'sum_solid.compute' in step_events
    assert 'sum_sq_solid.compute' in step_events

    snapshot.assert_match(clean_log_messages(result.data))
    store = build_fs_intermediate_store(instance.intermediates_directory, pipeline_run.run_id)
    intermediates_manager = IntermediateStoreIntermediatesManager(store)
    assert not intermediates_manager.has_intermediate(None, StepOutputHandle('sum_solid.compute'))
    assert not intermediates_manager.has_intermediate(
        None, StepOutputHandle('sum_sq_solid.compute')
    )


def test_successful_one_part_execute_plan(graphql_context, snapshot):
    instance = graphql_context.instance
    run_config = csv_hello_world_solids_config_fs_storage()
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=csv_hello_world, run_config=run_config
    )
    selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')

    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            'executionParams': {
                'selector': selector,
                'runConfigData': run_config,
                'stepKeys': ['sum_solid.compute'],
                'executionMetadata': {'runId': pipeline_run.run_id},
                'mode': 'default',
            },
        },
    )

    query_result = result.data['executePlan']

    assert query_result['__typename'] == 'ExecutePlanSuccess'
    assert query_result['pipeline']['name'] == 'csv_hello_world'
    assert query_result['hasFailures'] is False

    step_events = query_result['stepEvents']

    assert [se['__typename'] for se in step_events] == [
        'ExecutionStepStartEvent',
        'ExecutionStepInputEvent',
        'ExecutionStepOutputEvent',
        'ObjectStoreOperationEvent',
        'ExecutionStepSuccessEvent',
    ]

    assert step_events[1]['stepKey'] == 'sum_solid.compute'
    assert step_events[2]['outputName'] == 'result'

    expected_value_repr = (
        '''[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3)]), '''
        '''OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7)])]'''
    )

    assert step_events[3]['stepKey'] == 'sum_solid.compute'
    assert step_events[4]['stepKey'] == 'sum_solid.compute'

    snapshot.assert_match(clean_log_messages(result.data))

    store = build_fs_intermediate_store(instance.intermediates_directory, pipeline_run.run_id)
    intermediates_manager = IntermediateStoreIntermediatesManager(store)
    assert intermediates_manager.has_intermediate(None, StepOutputHandle('sum_solid.compute'))
    assert (
        str(
            intermediates_manager.get_intermediate(
                None, PoorMansDataFrame, StepOutputHandle('sum_solid.compute')
            ).obj
        )
        == expected_value_repr
    )


def test_successful_two_part_execute_plan(graphql_context, snapshot):
    instance = graphql_context.instance
    run_config = csv_hello_world_solids_config_fs_storage()
    pipeline_run = instance.create_run_for_pipeline(
        pipeline_def=csv_hello_world, run_config=run_config
    )
    selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
    result_one = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            'executionParams': {
                'selector': selector,
                'runConfigData': run_config,
                'stepKeys': ['sum_solid.compute'],
                'executionMetadata': {'runId': pipeline_run.run_id},
                'mode': 'default',
            },
        },
    )

    assert result_one.data['executePlan']['__typename'] == 'ExecutePlanSuccess'

    snapshot.assert_match(clean_log_messages(result_one.data))

    result_two = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            'executionParams': {
                'selector': selector,
                'runConfigData': csv_hello_world_solids_config_fs_storage(),
                'stepKeys': ['sum_sq_solid.compute'],
                'executionMetadata': {'runId': pipeline_run.run_id},
                'mode': 'default',
            },
        },
    )

    query_result = result_two.data['executePlan']
    assert query_result['__typename'] == 'ExecutePlanSuccess'
    assert query_result['pipeline']['name'] == 'csv_hello_world'
    assert query_result['hasFailures'] is False
    step_events = query_result['stepEvents']
    assert [se['__typename'] for se in step_events] == [
        'ExecutionStepStartEvent',
        'ObjectStoreOperationEvent',
        'ExecutionStepInputEvent',
        'ExecutionStepOutputEvent',
        'ObjectStoreOperationEvent',
        'ExecutionStepSuccessEvent',
    ]
    assert step_events[0]['stepKey'] == 'sum_sq_solid.compute'
    assert step_events[1]['stepKey'] == 'sum_sq_solid.compute'
    assert step_events[2]['stepKey'] == 'sum_sq_solid.compute'
    assert step_events[3]['outputName'] == 'result'
    assert step_events[4]['stepKey'] == 'sum_sq_solid.compute'

    snapshot.assert_match(clean_log_messages(result_two.data))

    expected_value_repr = (
        '''[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3), '''
        '''('sum_sq', 9)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7), '''
        '''('sum_sq', 49)])]'''
    )

    store = build_fs_intermediate_store(instance.intermediates_directory, pipeline_run.run_id)

    intermediates_manager = IntermediateStoreIntermediatesManager(store)
    assert intermediates_manager.has_intermediate(None, StepOutputHandle('sum_sq_solid.compute'))
    assert (
        str(
            intermediates_manager.get_intermediate(
                None, PoorMansDataFrame, StepOutputHandle('sum_sq_solid.compute')
            ).obj
        )
        == expected_value_repr
    )


class TestExecutionPlan(ReadonlyGraphQLContextTestMatrix):
    def test_invalid_config_fetch_execute_plan(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        result = execute_dagster_graphql(
            graphql_context,
            EXECUTION_PLAN_QUERY,
            variables={
                'pipeline': selector,
                'runConfigData': {
                    'solids': {'sum_solid': {'inputs': {'num': {'csv': {'path': 384938439}}}}}
                },
                'mode': 'default',
            },
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['executionPlanOrError']['__typename'] == 'PipelineConfigValidationInvalid'
        )
        assert len(result.data['executionPlanOrError']['errors']) == 1
        assert (
            'Invalid scalar at path root:solids:sum_solid:inputs:num'
            in result.data['executionPlanOrError']['errors'][0]['message']
        )
        result.data['executionPlanOrError']['errors'][0][
            'message'
        ] = 'Invalid scalar at path root:solids:sum_solid:inputs:num'
        snapshot.assert_match(result.data)


def test_invalid_config_execute_plan(graphql_context, snapshot):
    selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            'executionParams': {
                'selector': selector,
                'runConfigData': {
                    'solids': {'sum_solid': {'inputs': {'num': {'csv': {'path': 384938439}}}}}
                },
                'stepKeys': [
                    'sum_solid.num.input_thunk',
                    'sum_solid.compute',
                    'sum_sq_solid.compute',
                ],
                'executionMetadata': {'runId': 'kdjkfjdfd'},
                'mode': 'default',
            },
        },
    )

    assert not result.errors
    assert result.data
    assert result.data['executePlan']['__typename'] == 'PipelineConfigValidationInvalid'
    assert len(result.data['executePlan']['errors']) == 1
    assert (
        'Invalid scalar at path root:solids:sum_solid:inputs:num'
        in result.data['executePlan']['errors'][0]['message']
    )
    result.data['executePlan']['errors'][0][
        'message'
    ] = 'Invalid scalar at path root:solids:sum_solid:inputs:num'
    snapshot.assert_match(result.data)


def test_pipeline_not_found_error_execute_plan(graphql_context, snapshot):
    selector = infer_pipeline_selector(graphql_context, 'nope')
    result = execute_dagster_graphql(
        graphql_context,
        EXECUTE_PLAN_QUERY,
        variables={
            'executionParams': {
                'selector': selector,
                'runConfigData': {
                    'solids': {'sum_solid': {'inputs': {'num': {'csv': {'path': 'ok'}}}}}
                },
                'stepKeys': [
                    'sum_solid.num.input_thunk',
                    'sum_solid.compute',
                    'sum_sq_solid.compute',
                ],
                'executionMetadata': {'runId': 'kdjkfjdfd'},
                'mode': 'default',
            },
        },
    )

    assert result.data['executePlan']['__typename'] == 'PipelineNotFoundError'
    assert result.data['executePlan']['pipelineName'] == 'nope'
    snapshot.assert_match(result.data)


def test_basic_execute_plan_with_materialization(graphql_context):
    selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
    with get_temp_file_name() as out_csv_path:

        run_config = {
            'solids': {
                'sum_solid': {
                    'inputs': {'num': file_relative_path(__file__, '../data/num.csv')},
                    'outputs': [{'result': out_csv_path}],
                }
            }
        }

        result = execute_dagster_graphql(
            graphql_context,
            EXECUTION_PLAN_QUERY,
            variables={'pipeline': selector, 'runConfigData': run_config, 'mode': 'default',},
        )

        steps_data = result.data['executionPlanOrError']['steps']

        assert set([step_data['key'] for step_data in steps_data]) == set(
            ['sum_solid.compute', 'sum_sq_solid.compute',]
        )

        instance = graphql_context.instance

        pipeline_run = instance.create_run_for_pipeline(
            pipeline_def=csv_hello_world, run_config=run_config
        )

        result = execute_dagster_graphql(
            graphql_context,
            EXECUTE_PLAN_QUERY,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': run_config,
                    'stepKeys': ['sum_solid.compute', 'sum_sq_solid.compute'],
                    'executionMetadata': {'runId': pipeline_run.run_id},
                    'mode': 'default',
                },
            },
        )

        assert result.data

        step_mat_event = None

        for message in result.data['executePlan']['stepEvents']:
            if message['__typename'] == 'StepMaterializationEvent':
                # ensure only one event
                assert step_mat_event is None
                step_mat_event = message

        # ensure only one event
        assert step_mat_event
        assert step_mat_event['materialization']
        assert len(step_mat_event['materialization']['metadataEntries']) == 1
        metadata_entry = step_mat_event['materialization']['metadataEntries'][0]
        assert metadata_entry['path'] == out_csv_path



from time import sleep

from dagster_graphql.client.query import (
    LAUNCH_PIPELINE_EXECUTION_MUTATION,
    LAUNCH_PIPELINE_REEXECUTION_MUTATION,
    PIPELINE_REEXECUTION_INFO_QUERY,
)
from dagster_graphql.test.utils import (
    execute_dagster_graphql,
    execute_dagster_graphql_and_finish_runs,
    infer_pipeline_selector,
)

from dagster import DagsterEventType
from dagster.core.execution.plan.objects import StepOutputHandle
from dagster.core.storage.intermediate_store import build_fs_intermediate_store
from dagster.core.storage.intermediates_manager import IntermediateStoreIntermediatesManager
from dagster.core.storage.tags import RESUME_RETRY_TAG
from dagster.core.utils import make_new_run_id

from .graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
    GraphQLContextVariant,
    OutOfProcessExecutingGraphQLContextTestMatrix,
    make_graphql_context_test_suite,
)
from .setup import (
    PoorMansDataFrame,
    csv_hello_world_solids_config,
    csv_hello_world_solids_config_fs_storage,
    get_retry_multi_execution_params,
    retry_config,
)
from .utils import get_all_logs_for_finished_run_via_subscription, sync_execute_get_events


def step_started(logs, step_key):
    return any(
        log['stepKey'] == step_key
        for log in logs
        if log['__typename'] in ('ExecutionStepStartEvent',)
    )


def step_did_not_run(logs, step_key):
    return not any(
        log['stepKey'] == step_key
        for log in logs
        if log['__typename']
        in ('ExecutionStepSuccessEvent', 'ExecutionStepSkippedEvent', 'ExecutionStepFailureEvent')
    )


def step_did_succeed(logs, step_key):
    return any(
        log['__typename'] == 'ExecutionStepSuccessEvent' and step_key == log['stepKey']
        for log in logs
    )


def step_did_skip(logs, step_key):
    return any(
        log['__typename'] == 'ExecutionStepSkippedEvent' and step_key == log['stepKey']
        for log in logs
    )


def step_did_fail(logs, step_key):
    return any(
        log['__typename'] == 'ExecutionStepFailureEvent' and step_key == log['stepKey']
        for log in logs
    )


def step_did_fail_in_records(records, step_key):
    return any(
        record.step_key == step_key
        and record.dagster_event.event_type_value == DagsterEventType.STEP_FAILURE.value
        for record in records
    )


def step_did_succeed_in_records(records, step_key):
    return any(
        record.step_key == step_key
        and record.dagster_event.event_type_value == DagsterEventType.STEP_SUCCESS.value
        for record in records
    )


def step_did_not_run_in_records(records, step_key):
    return not any(
        record.step_key == step_key
        and record.dagster_event.event_type_value
        in (
            DagsterEventType.STEP_SUCCESS.value,
            DagsterEventType.STEP_FAILURE.value,
            DagsterEventType.STEP_SKIPPED.value,
        )
        for record in records
    )


def first_event_of_type(logs, message_type):
    for log in logs:
        if log['__typename'] == message_type:
            return log
    return None


def has_event_of_type(logs, message_type):
    return first_event_of_type(logs, message_type) is not None


def get_step_output_event(logs, step_key, output_name='result'):
    for log in logs:
        if (
            log['__typename'] == 'ExecutionStepOutputEvent'
            and log['stepKey'] == step_key
            and log['outputName'] == output_name
        ):
            return log

    return None


class TestRetryExecution(ExecutingGraphQLContextTestMatrix):
    def test_retry_pipeline_execution(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'eventually_successful')
        result = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'mode': 'default',
                    'selector': selector,
                    'runConfigData': retry_config(0),
                }
            },
        )

        run_id = result.data['launchPipelineExecution']['run']['runId']
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            'pipelineRunLogs'
        ]['messages']

        assert step_did_succeed(logs, 'spawn.compute')
        assert step_did_fail(logs, 'fail.compute')
        assert step_did_skip(logs, 'fail_2.compute')
        assert step_did_skip(logs, 'fail_3.compute')
        assert step_did_skip(logs, 'reset.compute')

        retry_one = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'mode': 'default',
                    'selector': selector,
                    'runConfigData': retry_config(1),
                    'executionMetadata': {
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                        'tags': [{'key': RESUME_RETRY_TAG, 'value': 'true'}],
                    },
                }
            },
        )

        run_id = retry_one.data['launchPipelineReexecution']['run']['runId']
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            'pipelineRunLogs'
        ]['messages']
        assert step_did_not_run(logs, 'spawn.compute')
        assert step_did_succeed(logs, 'fail.compute')
        assert step_did_fail(logs, 'fail_2.compute')
        assert step_did_skip(logs, 'fail_3.compute')
        assert step_did_skip(logs, 'reset.compute')

        retry_two = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'mode': 'default',
                    'selector': selector,
                    'runConfigData': retry_config(2),
                    'executionMetadata': {
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                        'tags': [{'key': RESUME_RETRY_TAG, 'value': 'true'}],
                    },
                }
            },
        )

        run_id = retry_two.data['launchPipelineReexecution']['run']['runId']
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            'pipelineRunLogs'
        ]['messages']

        assert step_did_not_run(logs, 'spawn.compute')
        assert step_did_not_run(logs, 'fail.compute')
        assert step_did_succeed(logs, 'fail_2.compute')
        assert step_did_fail(logs, 'fail_3.compute')
        assert step_did_skip(logs, 'reset.compute')

        retry_three = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'mode': 'default',
                    'selector': selector,
                    'runConfigData': retry_config(3),
                    'executionMetadata': {
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                        'tags': [{'key': RESUME_RETRY_TAG, 'value': 'true'}],
                    },
                }
            },
        )

        run_id = retry_three.data['launchPipelineReexecution']['run']['runId']
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            'pipelineRunLogs'
        ]['messages']

        assert step_did_not_run(logs, 'spawn.compute')
        assert step_did_not_run(logs, 'fail.compute')
        assert step_did_not_run(logs, 'fail_2.compute')
        assert step_did_succeed(logs, 'fail_3.compute')
        assert step_did_succeed(logs, 'reset.compute')

    def test_retry_resource_pipeline(self, graphql_context):
        context = graphql_context
        selector = infer_pipeline_selector(graphql_context, 'retry_resource_pipeline')
        result = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'mode': 'default',
                    'selector': selector,
                    'runConfigData': {'storage': {'filesystem': {}}},
                }
            },
        )

        run_id = result.data['launchPipelineExecution']['run']['runId']
        logs = get_all_logs_for_finished_run_via_subscription(context, run_id)['pipelineRunLogs'][
            'messages'
        ]
        assert step_did_succeed(logs, 'start.compute')
        assert step_did_fail(logs, 'will_fail.compute')

        retry_one = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'mode': 'default',
                    'selector': selector,
                    'runConfigData': {'storage': {'filesystem': {}}},
                    'executionMetadata': {
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                        'tags': [{'key': RESUME_RETRY_TAG, 'value': 'true'}],
                    },
                }
            },
        )
        run_id = retry_one.data['launchPipelineReexecution']['run']['runId']
        logs = get_all_logs_for_finished_run_via_subscription(context, run_id)['pipelineRunLogs'][
            'messages'
        ]
        assert step_did_not_run(logs, 'start.compute')
        assert step_did_fail(logs, 'will_fail.compute')

    def test_retry_multi_output(self, graphql_context):
        context = graphql_context
        result = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': get_retry_multi_execution_params(context, should_fail=True)
            },
        )

        run_id = result.data['launchPipelineExecution']['run']['runId']
        logs = get_all_logs_for_finished_run_via_subscription(context, run_id)['pipelineRunLogs'][
            'messages'
        ]
        assert step_did_succeed(logs, 'multi.compute')
        assert step_did_skip(logs, 'child_multi_skip.compute')
        assert step_did_fail(logs, 'can_fail.compute')
        assert step_did_skip(logs, 'child_fail.compute')
        assert step_did_skip(logs, 'child_skip.compute')
        assert step_did_skip(logs, 'grandchild_fail.compute')

        retry_one = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': get_retry_multi_execution_params(
                    context, should_fail=True, retry_id=run_id
                )
            },
        )

        run_id = retry_one.data['launchPipelineReexecution']['run']['runId']
        logs = get_all_logs_for_finished_run_via_subscription(context, run_id)['pipelineRunLogs'][
            'messages'
        ]
        assert step_did_not_run(logs, 'multi.compute')
        assert step_did_not_run(logs, 'child_multi_skip.compute')
        assert step_did_fail(logs, 'can_fail.compute')
        assert step_did_skip(logs, 'child_fail.compute')
        assert step_did_skip(logs, 'child_skip.compute')
        assert step_did_skip(logs, 'grandchild_fail.compute')

        retry_two = execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': get_retry_multi_execution_params(
                    context, should_fail=False, retry_id=run_id
                )
            },
        )

        run_id = retry_two.data['launchPipelineReexecution']['run']['runId']
        logs = get_all_logs_for_finished_run_via_subscription(context, run_id)['pipelineRunLogs'][
            'messages'
        ]
        assert step_did_not_run(logs, 'multi.compute')
        assert step_did_not_run(logs, 'child_multi_skip.compute')
        assert step_did_succeed(logs, 'can_fail.compute')
        assert step_did_succeed(logs, 'child_fail.compute')
        assert step_did_skip(logs, 'child_skip.compute')
        assert step_did_succeed(logs, 'grandchild_fail.compute')

    def test_successful_pipeline_reexecution(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        run_id = make_new_run_id()
        result_one = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config_fs_storage(),
                    'executionMetadata': {'runId': run_id},
                    'mode': 'default',
                }
            },
        )

        assert (
            result_one.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineRunSuccess'
        )

        expected_value_repr = (
            '''[OrderedDict([('num1', '1'), ('num2', '2'), ('sum', 3), '''
            '''('sum_sq', 9)]), OrderedDict([('num1', '3'), ('num2', '4'), ('sum', 7), '''
            '''('sum_sq', 49)])]'''
        )

        instance = graphql_context.instance

        store = build_fs_intermediate_store(instance.intermediates_directory, run_id)
        intermediates_manager = IntermediateStoreIntermediatesManager(store)
        assert intermediates_manager.has_intermediate(None, StepOutputHandle('sum_solid.compute'))
        assert intermediates_manager.has_intermediate(
            None, StepOutputHandle('sum_sq_solid.compute')
        )
        assert (
            str(
                intermediates_manager.get_intermediate(
                    None, PoorMansDataFrame, StepOutputHandle('sum_sq_solid.compute')
                ).obj
            )
            == expected_value_repr
        )

        # retry
        new_run_id = make_new_run_id()

        result_two = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config_fs_storage(),
                    'stepKeys': ['sum_sq_solid.compute'],
                    'executionMetadata': {
                        'runId': new_run_id,
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                        'tags': [{'key': RESUME_RETRY_TAG, 'value': 'true'}],
                    },
                    'mode': 'default',
                }
            },
        )

        query_result = result_two.data['launchPipelineReexecution']
        assert query_result['__typename'] == 'LaunchPipelineRunSuccess'

        result = get_all_logs_for_finished_run_via_subscription(graphql_context, new_run_id)
        logs = result['pipelineRunLogs']['messages']

        assert isinstance(logs, list)
        assert has_event_of_type(logs, 'PipelineStartEvent')
        assert has_event_of_type(logs, 'PipelineSuccessEvent')
        assert not has_event_of_type(logs, 'PipelineFailureEvent')

        assert not get_step_output_event(logs, 'sum_solid.compute')
        assert get_step_output_event(logs, 'sum_sq_solid.compute')

        store = build_fs_intermediate_store(instance.intermediates_directory, new_run_id)
        intermediates_manager = IntermediateStoreIntermediatesManager(store)
        assert not intermediates_manager.has_intermediate(
            None, StepOutputHandle('sum_solid.inputs.num.read', 'input_thunk_output')
        )
        assert intermediates_manager.has_intermediate(None, StepOutputHandle('sum_solid.compute'))
        assert intermediates_manager.has_intermediate(
            None, StepOutputHandle('sum_sq_solid.compute')
        )
        assert (
            str(
                intermediates_manager.get_intermediate(
                    None, PoorMansDataFrame, StepOutputHandle('sum_sq_solid.compute')
                ).obj
            )
            == expected_value_repr
        )

    def test_pipeline_reexecution_info_query(self, graphql_context, snapshot):
        context = graphql_context
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')

        run_id = make_new_run_id()
        execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config_fs_storage(),
                    'executionMetadata': {'runId': run_id},
                    'mode': 'default',
                }
            },
        )

        # retry
        new_run_id = make_new_run_id()
        execute_dagster_graphql_and_finish_runs(
            context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config_fs_storage(),
                    'stepKeys': ['sum_sq_solid.compute'],
                    'executionMetadata': {
                        'runId': new_run_id,
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                        'tags': [{'key': RESUME_RETRY_TAG, 'value': 'true'}],
                    },
                    'mode': 'default',
                }
            },
        )

        result_one = execute_dagster_graphql_and_finish_runs(
            context, PIPELINE_REEXECUTION_INFO_QUERY, variables={'runId': run_id}
        )
        query_result_one = result_one.data['pipelineRunOrError']
        assert query_result_one['__typename'] == 'PipelineRun'
        assert query_result_one['stepKeysToExecute'] is None

        result_two = execute_dagster_graphql_and_finish_runs(
            context, PIPELINE_REEXECUTION_INFO_QUERY, variables={'runId': new_run_id}
        )
        query_result_two = result_two.data['pipelineRunOrError']
        assert query_result_two['__typename'] == 'PipelineRun'
        stepKeysToExecute = query_result_two['stepKeysToExecute']
        assert stepKeysToExecute is not None
        snapshot.assert_match(stepKeysToExecute)

    def test_pipeline_reexecution_invalid_step_in_subset(self, graphql_context):
        run_id = make_new_run_id()
        selector = infer_pipeline_selector(graphql_context, 'csv_hello_world')
        execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config(),
                    'executionMetadata': {'runId': run_id},
                    'mode': 'default',
                }
            },
        )

        # retry
        new_run_id = make_new_run_id()

        result_two = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'selector': selector,
                    'runConfigData': csv_hello_world_solids_config(),
                    'stepKeys': ['nope'],
                    'executionMetadata': {
                        'runId': new_run_id,
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                        'tags': [{'key': RESUME_RETRY_TAG, 'value': 'true'}],
                    },
                    'mode': 'default',
                }
            },
        )

        query_result = result_two.data['launchPipelineReexecution']
        assert query_result['__typename'] == 'InvalidStepError'
        assert query_result['invalidStepKey'] == 'nope'


class TestHardFailures(OutOfProcessExecutingGraphQLContextTestMatrix):
    def test_retry_hard_failure(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, 'hard_failer')
        result = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'mode': 'default',
                    'selector': selector,
                    'runConfigData': {'solids': {'hard_fail_or_0': {'config': {'fail': True}}}},
                }
            },
        )

        run_id = result.data['launchPipelineExecution']['run']['runId']
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            'pipelineRunLogs'
        ]['messages']

        assert step_started(logs, 'hard_fail_or_0.compute')
        assert step_did_not_run(logs, 'hard_fail_or_0.compute')
        assert step_did_not_run(logs, 'increment.compute')

        retry = execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'mode': 'default',
                    'selector': selector,
                    'runConfigData': {'solids': {'hard_fail_or_0': {'config': {'fail': False}}}},
                    'executionMetadata': {
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                        'tags': [{'key': RESUME_RETRY_TAG, 'value': 'true'}],
                    },
                }
            },
        )

        run_id = retry.data['launchPipelineReexecution']['run']['runId']
        logs = get_all_logs_for_finished_run_via_subscription(graphql_context, run_id)[
            'pipelineRunLogs'
        ]['messages']
        assert step_did_succeed(logs, 'hard_fail_or_0.compute')
        assert step_did_succeed(logs, 'increment.compute')


def _do_retry_intermediates_test(graphql_context, run_id, reexecution_run_id):
    selector = infer_pipeline_selector(graphql_context, 'eventually_successful')
    logs = sync_execute_get_events(
        context=graphql_context,
        variables={
            'executionParams': {
                'mode': 'default',
                'selector': selector,
                'executionMetadata': {'runId': run_id},
            }
        },
    )

    assert step_did_succeed(logs, 'spawn.compute')
    assert step_did_fail(logs, 'fail.compute')
    assert step_did_skip(logs, 'fail_2.compute')
    assert step_did_skip(logs, 'fail_3.compute')
    assert step_did_skip(logs, 'reset.compute')

    retry_one = execute_dagster_graphql_and_finish_runs(
        graphql_context,
        LAUNCH_PIPELINE_REEXECUTION_MUTATION,
        variables={
            'executionParams': {
                'mode': 'default',
                'selector': selector,
                'executionMetadata': {
                    'runId': reexecution_run_id,
                    'rootRunId': run_id,
                    'parentRunId': run_id,
                    'tags': [{'key': RESUME_RETRY_TAG, 'value': 'true'}],
                },
            }
        },
    )

    return retry_one


class TestRetryExecutionSyncOnlyBehavior(
    make_graphql_context_test_suite(
        context_variants=[GraphQLContextVariant.in_memory_instance_in_process_env()]
    )
):
    def test_retry_requires_intermediates_sync_only(self, graphql_context):
        run_id = make_new_run_id()
        reexecution_run_id = make_new_run_id()

        retry_one = _do_retry_intermediates_test(graphql_context, run_id, reexecution_run_id)
        assert not retry_one.errors
        assert retry_one.data
        assert retry_one.data['launchPipelineReexecution']['__typename'] == 'PythonError'
        assert (
            'Cannot perform reexecution with non persistent intermediates manager'
            in retry_one.data['launchPipelineReexecution']['message']
        )


class TestRetryExecutionAsyncOnlyBehavior(
    make_graphql_context_test_suite(
        context_variants=[GraphQLContextVariant.sqlite_with_cli_api_run_launcher_in_process_env()]
    )
):
    def test_retry_requires_intermediates_async_only(self, graphql_context):
        run_id = make_new_run_id()
        reexecution_run_id = make_new_run_id()

        _do_retry_intermediates_test(graphql_context, run_id, reexecution_run_id)
        reexecution_run = graphql_context.instance.get_run_by_id(reexecution_run_id)

        assert reexecution_run.is_failure

    def test_retry_early_terminate(self, graphql_context):
        instance = graphql_context.instance
        selector = infer_pipeline_selector(
            graphql_context, 'retry_multi_input_early_terminate_pipeline'
        )
        run_id = make_new_run_id()
        execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'mode': 'default',
                    'selector': selector,
                    'runConfigData': {
                        'solids': {
                            'get_input_one': {'config': {'wait_to_terminate': True}},
                            'get_input_two': {'config': {'wait_to_terminate': True}},
                        },
                        'storage': {'filesystem': {}},
                    },
                    'executionMetadata': {'runId': run_id},
                }
            },
        )
        # Wait until the first step succeeded
        while instance.get_run_stats(run_id).steps_succeeded < 1:
            sleep(0.1)
        # Terminate the current pipeline run at the second step
        graphql_context.instance.run_launcher.terminate(run_id)

        records = instance.all_logs(run_id)

        # The first step should succeed, the second should fail or not start,
        # and the following steps should not appear in records
        assert step_did_succeed_in_records(records, 'return_one.compute')
        assert any(
            [
                step_did_fail_in_records(records, 'get_input_one.compute'),
                step_did_not_run_in_records(records, 'get_input_one.compute'),
            ]
        )
        assert step_did_not_run_in_records(records, 'get_input_two.compute')
        assert step_did_not_run_in_records(records, 'sum_inputs.compute')

        # Start retry
        new_run_id = make_new_run_id()

        execute_dagster_graphql_and_finish_runs(
            graphql_context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                'executionParams': {
                    'mode': 'default',
                    'selector': selector,
                    'runConfigData': {
                        'solids': {
                            'get_input_one': {'config': {'wait_to_terminate': False}},
                            'get_input_two': {'config': {'wait_to_terminate': False}},
                        },
                        'storage': {'filesystem': {}},
                    },
                    'executionMetadata': {
                        'runId': new_run_id,
                        'rootRunId': run_id,
                        'parentRunId': run_id,
                        'tags': [{'key': RESUME_RETRY_TAG, 'value': 'true'}],
                    },
                }
            },
        )

        retry_records = instance.all_logs(new_run_id)
        # The first step should not run and the other three steps should succeed in retry
        assert step_did_not_run_in_records(retry_records, 'return_one.compute')
        assert step_did_succeed_in_records(retry_records, 'get_input_one.compute')
        assert step_did_succeed_in_records(retry_records, 'get_input_two.compute')
        assert step_did_succeed_in_records(retry_records, 'sum_inputs.compute')



from dagster import DagsterInstance, seven
from dagster.core.launcher.sync_in_memory_run_launcher import SyncInMemoryRunLauncher
from dagster.core.test_utils import create_run_for_test

from .setup import get_main_external_repo


def test_sync_run_launcher_from_configurable_class():
    with seven.TemporaryDirectory() as temp_dir:
        instance_no_hijack = DagsterInstance.local_temp(
            temp_dir,
            overrides={
                'run_launcher': {
                    'module': 'dagster.core.launcher.sync_in_memory_run_launcher',
                    'class': 'SyncInMemoryRunLauncher',
                }
            },
        )

        assert isinstance(instance_no_hijack.run_launcher, SyncInMemoryRunLauncher)


def test_sync_run_launcher_run():
    external_repo = get_main_external_repo()
    external_pipeline = external_repo.get_full_external_pipeline('noop_pipeline')

    with seven.TemporaryDirectory() as temp_dir:
        instance = DagsterInstance.local_temp(
            temp_dir,
            overrides={
                'run_launcher': {
                    'module': 'dagster.core.launcher.sync_in_memory_run_launcher',
                    'class': 'SyncInMemoryRunLauncher',
                }
            },
        )

        run = create_run_for_test(instance=instance, pipeline_name=external_pipeline.name)

        run = instance.run_launcher.launch_run(
            instance=instance, run=run, external_pipeline=external_pipeline
        )

        completed_run = instance.get_run_by_id(run.run_id)
        assert completed_run.is_success



import csv
import datetime
import logging
import time
from collections import OrderedDict
from copy import deepcopy

from dagster_graphql.implementation.context import DagsterGraphQLContext
from dagster_graphql.test.utils import (
    define_context_for_file,
    define_out_of_process_context,
    infer_pipeline_selector,
)

from dagster import (
    Any,
    AssetMaterialization,
    Bool,
    DagsterInstance,
    Enum,
    EnumValue,
    EventMetadataEntry,
    ExpectationResult,
    Field,
    InputDefinition,
    Int,
    ModeDefinition,
    Noneable,
    Nothing,
    Output,
    OutputDefinition,
    Partition,
    PartitionSetDefinition,
    PresetDefinition,
    PythonObjectDagsterType,
    ScheduleDefinition,
    String,
    check,
    composite_solid,
    dagster_type_loader,
    dagster_type_materializer,
    daily_schedule,
    hourly_schedule,
    lambda_solid,
    logger,
    monthly_schedule,
    pipeline,
    repository,
    resource,
    solid,
    usable_as_dagster_type,
    weekly_schedule,
)
from dagster.core.definitions.partition import last_empty_partition
from dagster.core.definitions.reconstructable import ReconstructableRepository
from dagster.core.host_representation import InProcessRepositoryLocation
from dagster.core.log_manager import coerce_valid_log_level
from dagster.core.storage.tags import RESUME_RETRY_TAG
from dagster.utils import file_relative_path, segfault


@dagster_type_loader(String)
def df_input_schema(_context, path):
    with open(path, 'r') as fd:
        return [OrderedDict(sorted(x.items(), key=lambda x: x[0])) for x in csv.DictReader(fd)]


@dagster_type_materializer(String)
def df_output_schema(_context, path, value):
    with open(path, 'w') as fd:
        writer = csv.DictWriter(fd, fieldnames=value[0].keys())
        writer.writeheader()
        writer.writerows(rowdicts=value)

    return AssetMaterialization.file(path)


PoorMansDataFrame = PythonObjectDagsterType(
    python_type=list,
    name='PoorMansDataFrame',
    loader=df_input_schema,
    materializer=df_output_schema,
)


def define_test_out_of_process_context(instance):
    check.inst_param(instance, 'instance', DagsterInstance)
    return define_out_of_process_context(__file__, "test_repo", instance)


def define_test_context(instance):
    check.inst_param(instance, 'instance', DagsterInstance)
    return define_context_for_file(__file__, "test_repo", instance)


def create_main_recon_repo():
    return ReconstructableRepository.for_file(__file__, 'test_repo')


def get_main_external_repo():
    return InProcessRepositoryLocation(
        ReconstructableRepository.from_legacy_repository_yaml(
            file_relative_path(__file__, 'repo.yaml')
        ),
    ).get_repository('test_repo')


def define_test_snapshot_context():
    return DagsterGraphQLContext(
        instance=DagsterInstance.ephemeral(),
        locations=[InProcessRepositoryLocation(create_main_recon_repo())],
    )


def main_repo_location_name():
    return '<<in_process>>'


def main_repo_name():
    return 'test_repo'


@lambda_solid(
    input_defs=[InputDefinition('num', PoorMansDataFrame)],
    output_def=OutputDefinition(PoorMansDataFrame),
)
def sum_solid(num):
    sum_df = deepcopy(num)
    for x in sum_df:
        x['sum'] = int(x['num1']) + int(x['num2'])
    return sum_df


@lambda_solid(
    input_defs=[InputDefinition('sum_df', PoorMansDataFrame)],
    output_def=OutputDefinition(PoorMansDataFrame),
)
def sum_sq_solid(sum_df):
    sum_sq_df = deepcopy(sum_df)
    for x in sum_sq_df:
        x['sum_sq'] = int(x['sum']) ** 2
    return sum_sq_df


@solid(
    input_defs=[InputDefinition('sum_df', PoorMansDataFrame)],
    output_defs=[OutputDefinition(PoorMansDataFrame)],
)
def df_expectations_solid(_context, sum_df):
    yield ExpectationResult(label="some_expectation", success=True)
    yield ExpectationResult(label="other_expectation", success=True)
    yield Output(sum_df)


def csv_hello_world_solids_config():
    return {
        'solids': {
            'sum_solid': {'inputs': {'num': file_relative_path(__file__, '../data/num.csv')}}
        }
    }


def csv_hello_world_solids_config_fs_storage():
    return {
        'solids': {
            'sum_solid': {'inputs': {'num': file_relative_path(__file__, '../data/num.csv')}}
        },
        'storage': {'filesystem': {}},
    }


@solid(config_schema={'file': Field(String)})
def loop(context):
    with open(context.solid_config['file'], 'w') as ff:
        ff.write('yup')

    while True:
        time.sleep(0.1)


@pipeline
def infinite_loop_pipeline():
    loop()


@solid
def noop_solid(_):
    pass


@pipeline
def noop_pipeline():
    noop_solid()


@solid
def solid_asset_a(_):
    yield AssetMaterialization(asset_key='a')
    yield Output(1)


@solid
def solid_asset_b(_, num):
    yield AssetMaterialization(asset_key='b')
    time.sleep(0.1)
    yield AssetMaterialization(asset_key='c')
    yield Output(num)


@pipeline
def single_asset_pipeline():
    solid_asset_a()


@pipeline
def multi_asset_pipeline():
    solid_asset_b(solid_asset_a())


@pipeline
def pipeline_with_expectations():
    @solid(output_defs=[])
    def emit_successful_expectation(_context):
        yield ExpectationResult(
            success=True,
            label='always_true',
            description='Successful',
            metadata_entries=[
                EventMetadataEntry.json(label='data', data={'reason': 'Just because.'})
            ],
        )

    @solid(output_defs=[])
    def emit_failed_expectation(_context):
        yield ExpectationResult(
            success=False,
            label='always_false',
            description='Failure',
            metadata_entries=[
                EventMetadataEntry.json(label='data', data={'reason': 'Relentless pessimism.'})
            ],
        )

    @solid(output_defs=[])
    def emit_successful_expectation_no_metadata(_context):
        yield ExpectationResult(success=True, label='no_metadata', description='Successful')

    emit_successful_expectation()
    emit_failed_expectation()
    emit_successful_expectation_no_metadata()


@pipeline
def more_complicated_config():
    @solid(
        config_schema={
            'field_one': Field(String),
            'field_two': Field(String, is_required=False),
            'field_three': Field(String, is_required=False, default_value='some_value'),
        }
    )
    def a_solid_with_three_field_config(_context):
        return None

    noop_solid()
    a_solid_with_three_field_config()


@pipeline
def more_complicated_nested_config():
    @solid(
        name='a_solid_with_multilayered_config',
        input_defs=[],
        output_defs=[],
        config_schema={
            'field_any': Any,
            'field_one': String,
            'field_two': Field(String, is_required=False),
            'field_three': Field(String, is_required=False, default_value='some_value'),
            'nested_field': {
                'field_four_str': String,
                'field_five_int': Int,
                'field_six_nullable_int_list': Field([Noneable(int)], is_required=False),
            },
        },
    )
    def a_solid_with_multilayered_config(_):
        return None

    return a_solid_with_multilayered_config()


@pipeline(
    preset_defs=[
        PresetDefinition.from_files(
            name='prod',
            config_files=[
                file_relative_path(__file__, '../environments/csv_hello_world_prod.yaml')
            ],
        ),
        PresetDefinition.from_files(
            name='test',
            config_files=[
                file_relative_path(__file__, '../environments/csv_hello_world_test.yaml')
            ],
        ),
        PresetDefinition(
            name='test_inline',
            run_config={
                'solids': {
                    'sum_solid': {
                        'inputs': {'num': file_relative_path(__file__, '../data/num.csv')}
                    }
                }
            },
        ),
    ]
)
def csv_hello_world():
    return sum_sq_solid(sum_df=sum_solid())


@pipeline
def csv_hello_world_with_expectations():
    ss = sum_solid()
    sum_sq_solid(sum_df=ss)
    df_expectations_solid(sum_df=ss)


@pipeline
def csv_hello_world_two():
    return sum_solid()


@solid
def solid_that_gets_tags(context):
    return context.pipeline_run.tags


@pipeline(tags={'tag_key': 'tag_value'})
def hello_world_with_tags():
    return solid_that_gets_tags()


@solid(name='solid_with_list', input_defs=[], output_defs=[], config_schema=[int])
def solid_def(_):
    return None


@pipeline
def pipeline_with_list():
    solid_def()


@pipeline
def csv_hello_world_df_input():
    return sum_sq_solid(sum_solid())


@pipeline
def no_config_pipeline():
    @lambda_solid
    def return_hello():
        return 'Hello'

    return return_hello()


@pipeline
def no_config_chain_pipeline():
    @lambda_solid
    def return_foo():
        return 'foo'

    @lambda_solid
    def return_hello_world(_):
        return 'Hello World'

    return return_hello_world(return_foo())


@pipeline
def scalar_output_pipeline():
    @lambda_solid(output_def=OutputDefinition(String))
    def return_str():
        return 'foo'

    @lambda_solid(output_def=OutputDefinition(Int))
    def return_int():
        return 34234

    @lambda_solid(output_def=OutputDefinition(Bool))
    def return_bool():
        return True

    @lambda_solid(output_def=OutputDefinition(Any))
    def return_any():
        return 'dkjfkdjfe'

    return_str()
    return_int()
    return_bool()
    return_any()


@pipeline
def pipeline_with_enum_config():
    @solid(
        config_schema=Enum(
            'TestEnum',
            [
                EnumValue(config_value='ENUM_VALUE_ONE', description='An enum value.'),
                EnumValue(config_value='ENUM_VALUE_TWO', description='An enum value.'),
                EnumValue(config_value='ENUM_VALUE_THREE', description='An enum value.'),
            ],
        )
    )
    def takes_an_enum(_context):
        pass

    return takes_an_enum()


@pipeline
def naughty_programmer_pipeline():
    @lambda_solid
    def throw_a_thing():
        raise Exception('bad programmer, bad')

    return throw_a_thing()


@pipeline
def pipeline_with_invalid_definition_error():
    @usable_as_dagster_type(name='InputTypeWithoutHydration')
    class InputTypeWithoutHydration(int):
        pass

    @solid(output_defs=[OutputDefinition(InputTypeWithoutHydration)])
    def one(_):
        return 1

    @solid(
        input_defs=[InputDefinition('some_input', InputTypeWithoutHydration)],
        output_defs=[OutputDefinition(Int)],
    )
    def fail_subset(_, some_input):
        return some_input

    return fail_subset(one())


@resource(config_schema=Field(Int))
def adder_resource(init_context):
    return lambda x: x + init_context.resource_config


@resource(config_schema=Field(Int))
def multer_resource(init_context):
    return lambda x: x * init_context.resource_config


@resource(config_schema={'num_one': Field(Int), 'num_two': Field(Int)})
def double_adder_resource(init_context):
    return (
        lambda x: x
        + init_context.resource_config['num_one']
        + init_context.resource_config['num_two']
    )


@pipeline(
    mode_defs=[
        ModeDefinition(
            name='add_mode',
            resource_defs={'op': adder_resource},
            description='Mode that adds things',
        ),
        ModeDefinition(
            name='mult_mode',
            resource_defs={'op': multer_resource},
            description='Mode that multiplies things',
        ),
        ModeDefinition(
            name='double_adder',
            resource_defs={'op': double_adder_resource},
            description='Mode that adds two numbers to thing',
        ),
    ],
    preset_defs=[PresetDefinition.from_files("add", mode="add_mode")],
)
def multi_mode_with_resources():
    @solid(required_resource_keys={'op'})
    def apply_to_three(context):
        return context.resources.op(3)

    return apply_to_three()


@resource(config_schema=Field(Int, is_required=False))
def req_resource(_):
    return 1


@pipeline(mode_defs=[ModeDefinition(resource_defs={'R1': req_resource})])
def required_resource_pipeline():
    @solid(required_resource_keys={'R1'})
    def solid_with_required_resource(_):
        return 1

    solid_with_required_resource()


@logger(config_schema=Field(str))
def foo_logger(init_context):
    logger_ = logging.Logger('foo')
    logger_.setLevel(coerce_valid_log_level(init_context.logger_config))
    return logger_


@logger({'log_level': Field(str), 'prefix': Field(str)})
def bar_logger(init_context):
    class BarLogger(logging.Logger):
        def __init__(self, name, prefix, *args, **kwargs):
            self.prefix = prefix
            super(BarLogger, self).__init__(name, *args, **kwargs)

        def log(self, lvl, msg, *args, **kwargs):  # pylint: disable=arguments-differ
            msg = self.prefix + msg
            super(BarLogger, self).log(lvl, msg, *args, **kwargs)

    logger_ = BarLogger('bar', init_context.logger_config['prefix'])
    logger_.setLevel(coerce_valid_log_level(init_context.logger_config['log_level']))


@pipeline(
    mode_defs=[
        ModeDefinition(
            name='foo_mode', logger_defs={'foo': foo_logger}, description='Mode with foo logger'
        ),
        ModeDefinition(
            name='bar_mode', logger_defs={'bar': bar_logger}, description='Mode with bar logger'
        ),
        ModeDefinition(
            name='foobar_mode',
            logger_defs={'foo': foo_logger, 'bar': bar_logger},
            description='Mode with multiple loggers',
        ),
    ]
)
def multi_mode_with_loggers():
    @solid
    def return_six(context):
        context.log.critical('OMG!')
        return 6

    return return_six()


@pipeline
def composites_pipeline():
    @lambda_solid(input_defs=[InputDefinition('num', Int)], output_def=OutputDefinition(Int))
    def add_one(num):
        return num + 1

    @lambda_solid(input_defs=[InputDefinition('num')])
    def div_two(num):
        return num / 2

    @composite_solid(input_defs=[InputDefinition('num', Int)], output_defs=[OutputDefinition(Int)])
    def add_two(num):
        return add_one.alias('adder_2')(add_one.alias('adder_1')(num))

    @composite_solid(input_defs=[InputDefinition('num', Int)], output_defs=[OutputDefinition(Int)])
    def add_four(num):
        return add_two.alias('adder_2')(add_two.alias('adder_1')(num))

    @composite_solid
    def div_four(num):
        return div_two.alias('div_2')(div_two.alias('div_1')(num))

    return div_four(add_four())


@pipeline
def materialization_pipeline():
    @solid
    def materialize(_):
        yield AssetMaterialization(
            asset_key='all_types',
            description='a materialization with all metadata types',
            metadata_entries=[
                EventMetadataEntry.text('text is cool', 'text'),
                EventMetadataEntry.url('https://bigty.pe/neato', 'url'),
                EventMetadataEntry.fspath('/tmp/awesome', 'path'),
                EventMetadataEntry.json({'is_dope': True}, 'json'),
                EventMetadataEntry.python_artifact(EventMetadataEntry, 'python class'),
                EventMetadataEntry.python_artifact(file_relative_path, 'python function'),
                EventMetadataEntry.float(1.2, 'float'),
            ],
        )
        yield Output(None)

    materialize()


@pipeline
def spew_pipeline():
    @solid
    def spew(_):
        print('HELLO WORLD')  # pylint: disable=print-call

    spew()


def retry_config(count):
    return {
        'resources': {'retry_count': {'config': {'count': count}}},
        'storage': {'filesystem': {}},
    }


@resource(config_schema={'count': Field(Int, is_required=False, default_value=0)})
def retry_config_resource(context):
    return context.resource_config['count']


@pipeline(mode_defs=[ModeDefinition(resource_defs={'retry_count': retry_config_resource})])
def eventually_successful():
    @solid(output_defs=[OutputDefinition(Int)])
    def spawn(_):
        return 0

    @solid(
        input_defs=[InputDefinition('depth', Int)],
        output_defs=[OutputDefinition(Int)],
        required_resource_keys={'retry_count'},
    )
    def fail(context, depth):
        if context.resources.retry_count <= depth:
            raise Exception('fail')

        return depth + 1

    @solid
    def reset(_, depth):
        return depth

    reset(fail(fail(fail(spawn()))))


@pipeline
def hard_failer():
    @solid(
        config_schema={'fail': Field(Bool, is_required=False, default_value=False)},
        output_defs=[OutputDefinition(Int)],
    )
    def hard_fail_or_0(context):
        if context.solid_config['fail']:
            segfault()
        return 0

    @solid(input_defs=[InputDefinition('n', Int)],)
    def increment(_, n):
        return n + 1

    increment(hard_fail_or_0())


@resource
def resource_a(_):
    return 'A'


@resource
def resource_b(_):
    return 'B'


@solid(required_resource_keys={'a'})
def start(context):
    assert context.resources.a == 'A'
    return 1


@solid(required_resource_keys={'b'})
def will_fail(context, num):  # pylint: disable=unused-argument
    assert context.resources.b == 'B'
    raise Exception('fail')


@pipeline(mode_defs=[ModeDefinition(resource_defs={'a': resource_a, 'b': resource_b})])
def retry_resource_pipeline():
    will_fail(start())


@solid(
    config_schema={'fail': bool},
    input_defs=[InputDefinition('inp', str)],
    output_defs=[
        OutputDefinition(str, 'start_fail', is_required=False),
        OutputDefinition(str, 'start_skip', is_required=False),
    ],
)
def can_fail(context, inp):  # pylint: disable=unused-argument
    if context.solid_config['fail']:
        raise Exception('blah')

    yield Output('okay perfect', 'start_fail')


@solid(
    output_defs=[
        OutputDefinition(str, 'success', is_required=False),
        OutputDefinition(str, 'skip', is_required=False),
    ],
)
def multi(_):
    yield Output('okay perfect', 'success')


@solid
def passthrough(_, value):
    return value


@solid(input_defs=[InputDefinition('start', Nothing)], output_defs=[])
def no_output(_):
    yield ExpectationResult(True)


@pipeline
def retry_multi_output_pipeline():
    multi_success, multi_skip = multi()
    fail, skip = can_fail(multi_success)
    no_output.alias('child_multi_skip')(multi_skip)
    no_output.alias('child_skip')(skip)
    no_output.alias('grandchild_fail')(passthrough.alias('child_fail')(fail))


@pipeline(tags={'foo': 'bar'})
def tagged_pipeline():
    @lambda_solid
    def simple_solid():
        return 'Hello'

    return simple_solid()


@pipeline
def retry_multi_input_early_terminate_pipeline():
    @lambda_solid(output_def=OutputDefinition(Int))
    def return_one():
        return 1

    @solid(
        config_schema={'wait_to_terminate': bool},
        input_defs=[InputDefinition('one', Int)],
        output_defs=[OutputDefinition(Int)],
    )
    def get_input_one(context, one):
        if context.solid_config['wait_to_terminate']:
            while True:
                time.sleep(0.1)
        return one

    @solid(
        config_schema={'wait_to_terminate': bool},
        input_defs=[InputDefinition('one', Int)],
        output_defs=[OutputDefinition(Int)],
    )
    def get_input_two(context, one):
        if context.solid_config['wait_to_terminate']:
            while True:
                time.sleep(0.1)
        return one

    @lambda_solid(
        input_defs=[InputDefinition('input_one', Int), InputDefinition('input_two', Int)],
        output_def=OutputDefinition(Int),
    )
    def sum_inputs(input_one, input_two):
        return input_one + input_two

    step_one = return_one()
    return sum_inputs(input_one=get_input_one(step_one), input_two=get_input_two(step_one))


def get_retry_multi_execution_params(graphql_context, should_fail, retry_id=None):
    selector = infer_pipeline_selector(graphql_context, 'retry_multi_output_pipeline')
    return {
        'mode': 'default',
        'selector': selector,
        'runConfigData': {
            'storage': {'filesystem': {}},
            'solids': {'can_fail': {'config': {'fail': should_fail}}},
        },
        'executionMetadata': {
            'rootRunId': retry_id,
            'parentRunId': retry_id,
            'tags': [{'key': RESUME_RETRY_TAG, 'value': 'true'}],
        },
    }


def define_schedules():
    integer_partition_set = PartitionSetDefinition(
        name='scheduled_integer_partitions',
        pipeline_name='no_config_pipeline',
        partition_fn=lambda: [Partition(x) for x in range(1, 10)],
        run_config_fn_for_partition=lambda _partition: {"storage": {"filesystem": {}}},
        tags_fn_for_partition=lambda _partition: {"test": "1234"},
    )

    no_config_pipeline_hourly_schedule = ScheduleDefinition(
        name="no_config_pipeline_hourly_schedule",
        cron_schedule="0 0 * * *",
        pipeline_name="no_config_pipeline",
        run_config={"storage": {"filesystem": {}}},
    )

    no_config_pipeline_hourly_schedule_with_config_fn = ScheduleDefinition(
        name="no_config_pipeline_hourly_schedule_with_config_fn",
        cron_schedule="0 0 * * *",
        pipeline_name="no_config_pipeline",
        run_config_fn=lambda _context: {"storage": {"filesystem": {}}},
    )

    no_config_should_execute = ScheduleDefinition(
        name="no_config_should_execute",
        cron_schedule="0 0 * * *",
        pipeline_name="no_config_pipeline",
        run_config={"storage": {"filesystem": {}}},
        should_execute=lambda _context: False,
    )

    dynamic_config = ScheduleDefinition(
        name="dynamic_config",
        cron_schedule="0 0 * * *",
        pipeline_name="no_config_pipeline",
        run_config_fn=lambda _context: {"storage": {"filesystem": {}}},
    )

    partition_based = integer_partition_set.create_schedule_definition(
        schedule_name="partition_based", cron_schedule="0 0 * * *",
    )

    partition_based_custom_selector = integer_partition_set.create_schedule_definition(
        schedule_name="partition_based_custom_selector",
        cron_schedule="0 0 * * *",
        partition_selector=last_empty_partition,
    )

    @daily_schedule(
        pipeline_name='no_config_pipeline',
        start_date=datetime.datetime.now() - datetime.timedelta(days=1),
        execution_time=(datetime.datetime.now() + datetime.timedelta(hours=2)).time(),
    )
    def partition_based_decorator(_date):
        return {"storage": {"filesystem": {}}}

    @daily_schedule(
        pipeline_name='multi_mode_with_loggers',
        start_date=datetime.datetime.now() - datetime.timedelta(days=1),
        execution_time=(datetime.datetime.now() + datetime.timedelta(hours=2)).time(),
        mode='foo_mode',
    )
    def partition_based_multi_mode_decorator(_date):
        return {"storage": {"filesystem": {}}}

    @hourly_schedule(
        pipeline_name='no_config_chain_pipeline',
        start_date=datetime.datetime.now() - datetime.timedelta(days=1),
        execution_time=(datetime.datetime.now() + datetime.timedelta(hours=2)).time(),
        solid_selection=['return_foo'],
    )
    def solid_selection_hourly_decorator(_date):
        return {"storage": {"filesystem": {}}}

    @daily_schedule(
        pipeline_name='no_config_chain_pipeline',
        start_date=datetime.datetime.now() - datetime.timedelta(days=2),
        execution_time=(datetime.datetime.now() + datetime.timedelta(hours=3)).time(),
        solid_selection=['return_foo'],
    )
    def solid_selection_daily_decorator(_date):
        return {"storage": {"filesystem": {}}}

    @monthly_schedule(
        pipeline_name='no_config_chain_pipeline',
        start_date=datetime.datetime.now() - datetime.timedelta(days=100),
        execution_time=(datetime.datetime.now() + datetime.timedelta(hours=4)).time(),
        solid_selection=['return_foo'],
    )
    def solid_selection_monthly_decorator(_date):
        return {"storage": {"filesystem": {}}}

    @weekly_schedule(
        pipeline_name='no_config_chain_pipeline',
        start_date=datetime.datetime.now() - datetime.timedelta(days=50),
        execution_time=(datetime.datetime.now() + datetime.timedelta(hours=5)).time(),
        solid_selection=['return_foo'],
    )
    def solid_selection_weekly_decorator(_date):
        return {"storage": {"filesystem": {}}}

    # Schedules for testing the user error boundary
    @daily_schedule(
        pipeline_name='no_config_pipeline',
        start_date=datetime.datetime.now() - datetime.timedelta(days=1),
        should_execute=lambda _: asdf,  # pylint: disable=undefined-variable
    )
    def should_execute_error_schedule(_date):
        return {"storage": {"filesystem": {}}}

    @daily_schedule(
        pipeline_name='no_config_pipeline',
        start_date=datetime.datetime.now() - datetime.timedelta(days=1),
        tags_fn_for_date=lambda _: asdf,  # pylint: disable=undefined-variable
    )
    def tags_error_schedule(_date):
        return {"storage": {"filesystem": {}}}

    @daily_schedule(
        pipeline_name='no_config_pipeline',
        start_date=datetime.datetime.now() - datetime.timedelta(days=1),
    )
    def run_config_error_schedule(_date):
        return asdf  # pylint: disable=undefined-variable

    tagged_pipeline_schedule = ScheduleDefinition(
        name="tagged_pipeline_schedule",
        cron_schedule="0 0 * * *",
        pipeline_name="tagged_pipeline",
        run_config={"storage": {"filesystem": {}}},
    )

    tagged_pipeline_override_schedule = ScheduleDefinition(
        name="tagged_pipeline_override_schedule",
        cron_schedule="0 0 * * *",
        pipeline_name="tagged_pipeline",
        run_config={"storage": {"filesystem": {}}},
        tags={'foo': 'notbar'},
    )

    invalid_config_schedule = ScheduleDefinition(
        name="invalid_config_schedule",
        cron_schedule="0 0 * * *",
        pipeline_name="pipeline_with_enum_config",
        run_config={"solids": {"takes_an_enum": {'config': "invalid"}}},
    )

    return [
        run_config_error_schedule,
        no_config_pipeline_hourly_schedule,
        no_config_pipeline_hourly_schedule_with_config_fn,
        no_config_should_execute,
        dynamic_config,
        partition_based,
        partition_based_custom_selector,
        partition_based_decorator,
        partition_based_multi_mode_decorator,
        solid_selection_hourly_decorator,
        solid_selection_daily_decorator,
        solid_selection_monthly_decorator,
        solid_selection_weekly_decorator,
        should_execute_error_schedule,
        tagged_pipeline_schedule,
        tagged_pipeline_override_schedule,
        tags_error_schedule,
        invalid_config_schedule,
    ]


def define_partitions():
    integer_set = PartitionSetDefinition(
        name="integer_partition",
        pipeline_name="no_config_pipeline",
        solid_selection=['return_hello'],
        mode="default",
        partition_fn=lambda: [Partition(i) for i in range(10)],
        run_config_fn_for_partition=lambda _: {"storage": {"filesystem": {}}},
        tags_fn_for_partition=lambda partition: {'foo': partition.name},
    )

    enum_set = PartitionSetDefinition(
        name="enum_partition",
        pipeline_name="noop_pipeline",
        partition_fn=lambda: ["one", "two", "three"],
        run_config_fn_for_partition=lambda _: {"storage": {"filesystem": {}}},
    )

    return [integer_set, enum_set]


@repository
def empty_repo():
    return []


@repository
def test_repo():
    return (
        [
            composites_pipeline,
            csv_hello_world_df_input,
            csv_hello_world_two,
            csv_hello_world_with_expectations,
            csv_hello_world,
            eventually_successful,
            hard_failer,
            hello_world_with_tags,
            infinite_loop_pipeline,
            materialization_pipeline,
            more_complicated_config,
            more_complicated_nested_config,
            multi_asset_pipeline,
            multi_mode_with_loggers,
            multi_mode_with_resources,
            naughty_programmer_pipeline,
            no_config_chain_pipeline,
            no_config_pipeline,
            noop_pipeline,
            pipeline_with_enum_config,
            pipeline_with_expectations,
            pipeline_with_invalid_definition_error,
            pipeline_with_list,
            required_resource_pipeline,
            retry_multi_input_early_terminate_pipeline,
            retry_multi_output_pipeline,
            retry_resource_pipeline,
            scalar_output_pipeline,
            single_asset_pipeline,
            spew_pipeline,
            tagged_pipeline,
        ]
        + define_schedules()
        + define_partitions()
    )
